#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Scrambled Subsampled Randomized Fourier Transform (SSRFT).

PyTorch-compatible implementation of the SSRFT (as a matrix-free linear operator
and other utilities), from
`[TYUC2019, 3.2] <https://arxiv.org/abs/1902.08651>`_.
"""


import torch
import torch_dct as dct

from .linops import BaseRandomLinOp
from .utils import BadShapeError, NoFlatError, rademacher_flip, randperm


# ##############################################################################
# # SSRFT
# ##############################################################################
def ssrft(x, out_dims, seed=0b1110101001010101011, dct_norm="ortho"):
    r"""Right (forward) matrix multiplication of the SSRFT.

    This function implements a matrix-free, right-matmul operator of the
    Scrambled Subsampled Randomized Fourier Transform (SSRFT) for real-valued
    signals, from `[TYUC2019, 3.2] <https://arxiv.org/abs/1902.08651>`_.

    .. math::

      \text{SSRFT} = R\,\mathcal{F}\,\Pi\,\mathcal{F}\,\Pi'

    Where :math:`R` is a random index-picker, \mathcal{F} is a Discrete Cosine
    Transform, and :math:`\Pi, \Pi'` are random permutations.

    :param x: Vector to be projected, such that ``y = SSRFT @ x``
    :param out_dims: Dimensions of output ``y``, must be less than ``dim(x)``
    :param seed: Random seed
    """
    # make sure all sources of randomness are CPU, to ensure cross-device
    # consistency of the operator
    if len(x.shape) != 1:
        raise NoFlatError("Only flat tensors supported!")
    x_len = len(x)
    assert out_dims <= x_len, "Projection to larger dimensions not supported!"
    seeds = [seed + i for i in range(5)]
    # first scramble: permute, rademacher, and DCT
    perm1 = randperm(x_len, seed=seeds[0], device="cpu")
    x, rad1 = rademacher_flip(x[perm1], seed=seeds[1], inplace=False)
    del perm1, rad1
    x = dct.dct(x, norm=dct_norm)
    # second scramble: permute, rademacher and DCT
    perm2 = randperm(x_len, seed=seeds[2], device="cpu")
    x, rad2 = rademacher_flip(x[perm2], seeds[3], inplace=False)
    del perm2, rad2
    x = dct.dct(x, norm=dct_norm)
    # extract random indices and return
    out_idxs = randperm(x_len, seed=seeds[4], device="cpu")[:out_dims]
    x = x[out_idxs]
    return x


def ssrft_adjoint(x, out_dims, seed=0b1110101001010101011, dct_norm="ortho"):
    r"""Left (adjoint) matrix multiplication of the SSRFT.

    Adjoint operator of SSRFT, such that ``x @ SSRFT = y``. See :func:`.ssrft`
    for more details. Note the following implementation detail:

    * Permutations are orthogonal transforms
    * Rademacher transforms are also orthogonal (also diagonal and self-inverse)
    * DCT/DFT are also orthogonal transforms
    * The index-picker :math:`R` is a subset of rows of I.

    With orthogonal operators, transform and inverse are the same. Therefore,
    this adjoint operator takes the following form:

    .. math::

       \text{SSRFT}^T =& (R\,\mathcal{F}\,\Pi\,\mathcal{F}\,\Pi')^T \\
       =& \Pi'^T \, \mathcal{F}^T \, \Pi^T \, \mathcal{F}^T \, R^T \\
       =& \Pi'^{-1} \, \mathcal{F}^{-1} \, \Pi^{-1} \, \mathcal{F}^{-1} \, R^T

    So we can make use of the inverses, except for :math:`R^T`, which is a
    column-truncated identity, so we embed the entries picked by :math:`R` into
    the corresponding indices, and leave the rest as zeros.
    """
    # make sure all sources of randomness are CPU, to ensure cross-device
    # consistency of the operator
    if len(x.shape) != 1:
        raise NoFlatError("Only flat tensors supported!")
    x_len = len(x)
    assert (
        out_dims >= x_len
    ), "Backprojection into smaller dimensions not supported!"
    #
    seeds = [seed + i for i in range(5)]
    result = torch.zeros(
        out_dims,
        dtype=x.dtype,
    ).to(x.device)
    # first embed signal into original indices
    out_idxs = randperm(out_dims, seed=seeds[4], device="cpu")[:x_len]
    result[out_idxs] = x
    del x
    # then do the idct, followed by rademacher and inverse permutation
    result = dct.idct(result, norm=dct_norm)
    rademacher_flip(result, seeds[3], inplace=True)
    perm2_inv = randperm(out_dims, seed=seeds[2], device="cpu", inverse=True)
    result = result[perm2_inv]
    del perm2_inv
    # second inverse pass
    result = dct.idct(result, norm=dct_norm)
    rademacher_flip(result, seeds[1], inplace=True)
    perm1_inv = randperm(out_dims, seed=seeds[0], device="cpu", inverse=True)
    result = result[perm1_inv]
    #
    return result


class SSRFT(BaseRandomLinOp):
    """Scrambled Subsampled Randomized Fourier Transform (SSRFT).

    This class encapsulates the left- and right-SSRFT transforms into a single
    linear operator, which is deterministic for the same shape and seed
    (particularly, also across different torch devices).
    """

    def __init__(self, shape, seed=0b1110101001010101011):
        """:param shape: ``(height, width)`` of this linear operator."""
        super().__init__(shape, seed)
        h, w = shape
        if h > w:
            raise BadShapeError("Height > width not supported!")
        # :param scale: Ideally, ``1/l``, where ``l`` is the average diagonal
        #   value of the covmat ``A.T @ A``, where ``A`` is a FastJLT operator,
        #   so that ``l2norm(x)`` approximates ``l2norm(Ax)``.
        self.scale = NotImplemented

    def matmul(self, x):
        """Forward (right) matrix-vector multiplication ``SSRFT @ x``.

        See parent class for more details.
        """
        return ssrft(x, self.shape[0], seed=self.seed, dct_norm="ortho")

    def rmatmul(self, x):
        """Adjoint (left) matrix-vector multiplication ``x @ SSRFT``.

        See parent class for more details.
        """
        return ssrft_adjoint(x, self.shape[1], seed=self.seed, dct_norm="ortho")

    def get_row(self, idx, dtype, device):
        """Returns SSRFT[idx, :] via left-matmul with a one-hot vector."""
        in_buff = torch.zeros(self.shape[0], dtype=dtype, device=device)
        in_buff[idx] = 1
        return in_buff @ self



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""``skerch``: Sketched matrix decompositions for PyTorch.

This implementation is based on the following publications, referenced
throughout the documentation:

* `[TYUC2019] <https://arxiv.org/abs/1902.08651>`_: Joel A. Tropp, Alp
  Yurtsever, Madeleine Udell, and Volkan Cevher. 2019. *“Streaming Low-rank
  Matrix Approximation with an Application to Scientific Simulation”*. SIAM
  Journal on Scientific Computing 41 (4): A2430–63.
* `[BN2022] <https://arxiv.org/abs/2201.10684>`_: Robert A. Baston and Yuji
  Nakatsukasa. 2022. *“Stochastic diagonal estimation: probabilistic bounds and
  an improved algorithm”*.  CoRR abs/2201.10684.
"""


# ##############################################################################
# # GLOBALS
# ##############################################################################

# format strings to store measurements in HDF5 files.
LO_FMT = "leftouter_{}.h5"
RO_FMT = "rightouter_{}.h5"
INNER_FMT = "inner_{}.h5"



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""A-posteriori error bounds.

CLI plugin to provide a-posteriori error bounds for given number of measurements
and tolerance.
"""


from ..a_posteriori import a_posteriori_error_bounds


# ##############################################################################
# # ENTRY POINT
# ##############################################################################
def main(num_meas, rel_err):
    """See :func:`.a_posteriori.a_posteriori_error_bounds`."""
    result = a_posteriori_error_bounds(num_meas, rel_err)
    print(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""A-posteriori utilities for sketched decompositions.

Given a (potentially matrix-free) linear operator, and our obtained low-rank
decomposition, we would like to know how similar they are. In
`[TYUC2019] <https://arxiv.org/abs/1902.08651>`_, a method is presented to
estimate the Frobenius residual error. Furthermore, a probabilistic bound to
said estimator is presented, providing confidence bounds for the estimation.
More measures increase said confidence.

Finally, a "scree plot" method is also introduced, providing upper and lower
bounds to estimate the resudual Frobenius error as a function of the rank.
This can be used to estimate the actual rank of our original linear operator,
and to choose the rank of our recovery.

This module implements this functionality.
"""

import math

import torch

from .utils import gaussian_noise


################################################################################
# # A POSTERIORI ERROR ESTIMATION
# ##############################################################################
def a_posteriori_error_bounds(num_measurements, rel_err):
    """Probabilistic bounds for a-posteriori sketch error estimation.

    Retrieves probabilistic error bounds presented in
    `[TYUC2019, 6.4] <https://arxiv.org/abs/1902.08651>`_, see module docstring
    for details. Note that this function does not perform the error estimation,
    it merely informs about how noisy would be the estimation given the desired
    number of measurements.

    :param int num_measurements: How many Gaussian measurements will be
      performed for the a-posteriori error estimation. More measurements means
      tighter error bounds.
    :param float rel_err: A float between 0 and 1, indicating the relative
      error that we want to consider. If ``x`` is the actual error we want to
      estimate, and ``y`` is our estimation, this function will return the
      probability that ``y`` is in the range ``[(1-rel_err)x, (1+rel_err)x]``.
    :returns: Probabilities that, given the indicated number
      of measurements, the a-posteriori method yields values outside of
      ``actual_error * (1 +/- rel_err)``. Ideally, we want a sufficient number
      of measurements, such that the returned probabilities for small
      ``rel_err`` are themselves small (this means that the corresponding error
      estimation is tight).
    """
    assert 0 <= rel_err <= 1, "rel_err expected between 0 and 1"
    experr = math.exp(rel_err)
    meas_half = num_measurements / 2
    #
    lo_p = (experr * (1 - rel_err)) ** meas_half
    hi_p = (experr / (1 + rel_err)) ** (-meas_half)
    #
    result = {
        f"P(err<={1 - rel_err}x)": lo_p,
        f"P(err>={1 + rel_err}x)": hi_p,
    }
    return result


def a_posteriori_error(
    lop1,
    lop2,
    num_measurements,
    seed=0b1110101001010101011,
    dtype=torch.float64,
    device="cpu",
    adjoint=False,
):
    r"""A-posteriori sketch error estimation.

    This function implements the error estimation procedure presented in
    `[TYUC2019, 6] <https://arxiv.org/abs/1902.08651>`_.
    For that, it performs ``num_measurements`` gaussian vector
    multiplications on linear operators ``lop1`` and ``lop2``, and then
    compares the results.
    The :math:`\ell_2` error between the obtained measurements is then an
    estimation of the Frobenius error between ``lop1`` and ``lop2``.

    :param lop1: Linear operator with a ``shape=(h, w)`` attribute and
      implementing a left (``x @ mat``) or right ``(mat @ x)`` matmul op.
      It is assumed to be compatible with the given torch ``dtype``.
    :param lop2: See ``lop1``.
    :param adjoint: If true, left-matmul is used, otherwise right-matmul.
    :returns: A tuple ``((f1_mean, f2_mean, diff_mean), (f1, f2, diff))``,
      where the last 3 elements are lists with the ``L_2^2`` norms of each
      random measurement for the first matrix (f1), second matrix (f2) and
      difference between matrices (diff). The first 3 elements are the averages
      over all given measurements. The final estimate for the error between
      ``lop1`` and ``lop2`` is then ``diff_mean``. To get confidence bounds
      on this estimate, use ``a_posteriori_error_bounds``.
    """
    assert lop1.shape == lop2.shape, "Mismatching shapes!"
    h, w = lop1.shape
    #
    frob1_pow2, frob2_pow2, diff_pow2 = [], [], []
    for i in range(num_measurements):
        rand = gaussian_noise(
            h if adjoint else w,
            mean=0.0,
            std=1.0,
            seed=seed + i,
            dtype=dtype,
            device=device,
        )
        rand_np = None
        if adjoint:
            meas1 = rand @ lop1
            meas2 = rand @ lop2
        else:
            try:
                meas1 = lop1 @ rand
            except TypeError:
                rand_np = rand.cpu().numpy() if rand_np is None else rand_np
                meas1 = torch.from_numpy(lop1 @ rand_np)
            try:
                meas2 = lop2 @ rand
            except TypeError:
                rand_np = rand.cpu().numpy() if rand_np is None else rand_np
                meas2 = torch.from_numpy(lop2 @ rand_np)
        #
        frob1_pow2.append(sum(meas1**2).item())
        frob2_pow2.append(sum(meas2**2).item())
        diff_pow2.append(sum((meas1 - meas2) ** 2).item())
    #
    frob1_pow2_mean = sum(frob1_pow2) / num_measurements
    frob2_pow2_mean = sum(frob2_pow2) / num_measurements
    diff_pow2_mean = sum(diff_pow2) / num_measurements
    return (
        (frob1_pow2_mean, frob2_pow2_mean, diff_pow2_mean),
        (frob1_pow2, frob2_pow2, diff_pow2),
    )


# ##############################################################################
# # SCREE
# ##############################################################################
def scree_bounds(S, ori_frob, err_frob):
    """A-posteriori scree bounds for low-rank approximations.

    Reference: `[TYUC2019, 6.5] <https://arxiv.org/abs/1902.08651>`_.

    When we do a sketched eigen- or singular-decomposition of a linear operator,
    we don't generally know a-priori where is the effective rank of said
    operator. A scree analysis is an a-posteriori tool to estimate the smallest
    rank that captures a sufficient amount of the actual matrix.

    It does so by efficiently retrieving upper and lower bounds for the actual
    proportion of residual energy that remains, as we increase the rank of our
    estimated matrix. Typically we look for an initial sharp decay, followed
    by an "elbow" at a low point, after which residual energy stops decaying
    fast. The location of the elbow is a good indicator for the effective rank,
    but quantitative methods are also possible (see [TYUC2019]).

    :param S: Vector of the estimated spectrum/singular values, expected to
      contain entries in non-ascending magnitude.
    :param ori_frob: Estimate (or exact if available) ``frob_norm(M)``, where
      ``M`` is the original linop that we are decomposing. This quantity is
      estimated by :func:`.a_posteriori_error`, but note that it returns
      ``frob_norm^2``, while this function requires the Frobenius norm itself.
    :param err_frob: A-posteriori estimate of ``frob_norm(M - M')``, where
      ``M'`` is the recovery of ``M`` with singular values ``S``.

    Usage example::

      f1_pow2, f2_pow2, res_pow2 = a_posteriori_error(M, M', ...)
      S = get_singular_values(M')  # in non-ascending magnitude
      scree_lo, scree_hi = scree_bounds(S, f1_pow2**0.5, res_pow2**0.5)
      # plot the scree bounds to find an elbow as a cutting point for rank(M).
    """
    assert (
        abs(S).diff() <= 0
    ).all(), "Provided S must be given in non-ascending magnitude!"
    S_squared = S**2
    # residuals is a vector of len(S), where the ith entry contains the
    # Frobenius norm of M' after an ith-rank deflation, while progressing by
    # descending magnitude. The first entry has a rank 0 deflation, i.e. it
    # contains frob(M') entirely. The last entry has a rank R-1 deflation, i.e.
    # it is the magnitude of the smallest value in S.
    residuals = S_squared.flip(0).cumsum(0).flip(0) ** 0.5
    # now instead of normalizing dividing by sum(S_squared)**0.5, which uses M',
    # we obtain both the lower and upper bounds by using M instead, this is
    # given by ori_frob.
    lo_scree = (residuals / ori_frob) ** 2
    hi_scree = ((residuals + err_frob) / ori_frob) ** 2
    return lo_scree, hi_scree



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for a-priori functionality."""


import math

import pytest

from skerch.a_priori import a_priori_hyperparams


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def heights_widths_ranks_outer_inner_posteriori():
    """Test cases for different matrices and measurements.

    Entries are ``((h, w), r, (o, i))``. For a matrix of shape ``h, w`` and
    rank ``r``, the Skinny SVD will do ``o`` outer measurements and ``i``
    inner measurements. Recommended is that ``i >= 2*o``.

    For constrained budgets, the Skinny SVD will naturally yield higher error
    with smaller shapes, hence we test a few medium shapes.
    """
    result = [
        ((1_000, 1_000), 10, (100, 300), 30),
        ((1_000, 1_000), 50, (200, 600), 30),
        ((1_000, 2_000), 10, (100, 300), 30),
        ((1_000, 2_000), 50, (200, 600), 30),
        ((2_000, 2_000), 20, (100, 300), 30),
        ((2_000, 2_000), 100, (200, 600), 30),
    ]
    return result


@pytest.fixture
def budget_ratios():
    """Test cases for different memory budget to matrix size ratios."""
    result = [0.001, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    return result


# ##############################################################################
# # A PRIORI ASYMMETRIC HYPERPARAMETER ESTIMATION
# ##############################################################################
def check_outer_overflow(shape, outer, budget, hermitian=False):
    """Helper function.

    Test that increasing outer measurements by 1 breaks budget or constraints.
    """
    h, w = shape
    if hermitian:
        w = 0
    available = budget - ((outer + 1) * (h + w))
    assert available > 0, "Budget overflown!"
    # if not overflown, check if inner can be twice outer
    needed = (2 * (outer + 1)) ** 2
    assert available >= needed, "Inner measurements don't fit in budget!"


def test_a_priori_hyperparams(
    heights_widths_ranks_outer_inner_posteriori, budget_ratios
):
    """Test case for a-priori hyperparameters.

    The ``a_priori_hyperparams`` function provides suggestions for number of
    measurements, given a matrix shape and a budget. This function tests that:
    * suggested measurements are nonnegative
    * budget is respected
    * ``s >= 2k``, where ``s`` are core measurements and ``k`` the outer ones

    Ideally, incrementing k by 1 would explode or break constraints, meaning
    that ``k`` is indeed as large as it gets. This is tested by
    ``check_outer_overflow``, but it turns out this is not always the case.
    Since this is not crucial, we drop that test.
    """
    # test that symmetric requires square dimensions
    with pytest.raises(AssertionError):
        a_priori_hyperparams((10, 20), 50, complex_data=False, hermitian=True)
    #
    for (h, w), _r, (_o, _i), _ in heights_widths_ranks_outer_inner_posteriori:
        for ratio in budget_ratios:
            # test hpar recommendations for non-symmetric matrices
            budget = max(1, math.floor(h * w * ratio))
            oo, ii = a_priori_hyperparams(
                (h, w), budget, complex_data=False, hermitian=False
            )
            assert oo >= 0, "Outer measurements negative?"
            assert ii >= 0, "Inner measurements negative?"
            assert (oo + ii) > 0, "At least one measurement must be given!"
            assert ii >= (oo * 2), "Inner measurements must be >= 2*outer!"
            assert (oo * (h + w) + ii**2) <= budget, "Budget surpassed!"
            # It turns out some of the outputs can be incremented by 1 without
            # breaking budget or constraints. Maybe this is due to taking the
            # math.floor. Anyway, not super important, so we ignore it.
            # with pytest.raises(AssertionError):
            #     check_outer_overflow((h, w), oo, budget, hermitian=False)
            #
            # test hpar recommendations for symmetric matrices
            dim = min(h, w)
            oo, ii = a_priori_hyperparams(
                (dim, dim), budget, complex_data=False, hermitian=True
            )
            assert oo >= 0, "Outer measurements negative?"
            assert ii >= 0, "Inner measurements negative?"
            assert (oo + ii) > 0, "At least one measurement must be given!"
            assert ii >= (oo * 2), "Inner measurements must be >= 2*outer!"
            assert (oo * dim + ii**2) <= budget, "Budget surpassed!"
            # It turns out some of the outputs can be incremented by 1 without
            # breaking budget or constraints. Maybe this is due to taking the
            # math.floor. Anyway, not super important, so we ignore it.
            # with pytest.raises(AssertionError):
            #     check_outer_overflow((dim, dim), oo, budget, hermitian=True)


# ##############################################################################
# # A PRIORI SYMMETRIC HYPERPARAMETER ESTIMATION
# ##############################################################################



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytests distributed sketched SVD utils."""


import torch

from skerch.distributed_decompositions import (
    DistributedHDF5,
    create_hdf5_layout,
    orthogonalize,
    solve_core_ssvd,
)
from skerch.distributed_measurements import innermeas_idx_torch, ssrft_idx_torch
from skerch.ssrft import SSRFT


# ##############################################################################
# # MOCK DISTRIBUTED DECOMPOSITIONS
# ##############################################################################
def mock_dssvd(dirname, lop, lop_device, lop_dtype, outer, inner, seed):
    """Mock implementation of a distributed sketched SVD.

    This function emulates, in a single process, the steps that would be done
    in a distributed system to compute a sketched SVD. Usage example::

      tmpdir = tempfile.TemporaryDirectory()
      mock_dssvd(
          tmpdir.name, mat, device, dtype, outer, inner, seed
      )
      tmpdir.cleanup()
    """
    processing, success = "0", "1"  # arbitrary flags
    # convenience params
    h, w = lop.shape
    lo_seed, ro_seed, li_seed, ri_seed = seed, seed + 1, seed + 2, seed + 3
    # create HDF5 layout (centralized)
    (
        (lo_path, lo_subpaths),
        (ro_path, ro_subpaths),
        (inn_path, inn_subpaths),
    ) = create_hdf5_layout(
        dirname,
        (h, w),
        lop_dtype,
        outer,
        inner,
        lo_fmt="leftouter_{}.h5",
        ro_fmt="rightouter_{}.h5",
        inner_fmt="core_{}.h5",
        with_ro=True,
    )
    # perform left outer measurements (distributed)
    for i in range(outer):
        vals, flag, h5 = DistributedHDF5.load(lo_subpaths[i])
        ssrft_idx_torch(
            i,
            outer,
            lop,
            lop_device,
            lop_dtype,
            vals,
            lo_seed,
            flag,
            processing_flag=processing,
            success_flag=success,
            adjoint=True,
        )
        assert flag[0].decode() == success, f"Unsuccessful measurement? {i}"
        h5.close()

    # perform right outer measurements (distributed)
    for i in range(outer):
        vals, flag, h5 = DistributedHDF5.load(ro_subpaths[i])
        ssrft_idx_torch(
            i,
            outer,
            lop,
            lop_device,
            lop_dtype,
            vals,
            ro_seed,
            flag,
            processing_flag=processing,
            success_flag=success,
            adjoint=False,
        )
        assert flag[0].decode() == success, f"Unsuccessful measurement? {i}"
        h5.close()

    # perform inner measurements (distributed)
    for i in range(inner):
        vals, flag, h5 = DistributedHDF5.load(inn_subpaths[i])
        innermeas_idx_torch(
            i,
            inner,
            lop,
            lop_device,
            lop_dtype,
            vals,
            li_seed,
            ri_seed,
            flag=flag,
            processing_flag=processing,
            success_flag=success,
        )
        assert flag[0].decode() == success, f"Unsuccessful measurement? {i}"
        h5.close()

    # merge virtual HDF5 datasets into monolithic without memory overhead
    DistributedHDF5.merge_all(
        lo_path,
        out_path=None,
        delete_subfiles_while_merging=True,
        check_success_flag=success,
    )
    DistributedHDF5.merge_all(
        ro_path,
        out_path=None,
        delete_subfiles_while_merging=True,
        check_success_flag=success,
    )
    DistributedHDF5.merge_all(
        inn_path,
        out_path=None,
        delete_subfiles_while_merging=True,
        check_success_flag=success,
    )

    # load all components and check that all flags are OK (centralized)
    lo, lo_flags, lo_h5 = DistributedHDF5.load(lo_path, filemode="r+")
    ro, ro_flags, ro_h5 = DistributedHDF5.load(ro_path, filemode="r+")
    inn, inn_flags, inn_h5 = DistributedHDF5.load(inn_path, filemode="r+")
    assert all(f.decode() == "1" for f in lo_flags), "Bad left outer flags!"
    assert all(f.decode() == "1" for f in ro_flags), "Bad right outer flags!"
    assert all(f.decode() == "1" for f in inn_flags), "Bad inner flags!"
    # orthogonalize outer measurements (in-place)
    orthogonalize(lo, overwrite=True)
    orthogonalize(ro, overwrite=True)
    # Solve core op and decompose
    left_inner_ssrft = SSRFT((inner, h), seed=li_seed)
    right_inner_ssrft = SSRFT((inner, w), seed=ri_seed)
    core_U, core_S, core_Vt = solve_core_ssvd(
        ro, lo, inn, left_inner_ssrft, right_inner_ssrft
    )
    # for this utest, convert everything to explicit torch tensors
    ro = torch.from_numpy(ro[:]).to(lop_device)
    lo = torch.from_numpy(lo[:]).to(lop_device)
    core_U = torch.from_numpy(core_U).to(lop_device)
    core_S = torch.from_numpy(core_S).to(lop_device)
    core_Vt = torch.from_numpy(core_Vt).to(lop_device)
    # close HDF5 files and return
    lo_h5.close()
    ro_h5.close()
    inn_h5.close()
    return ro, core_U, core_S, core_Vt, lo.T



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for distributed measurement utils."""


import os
import tempfile

import numpy as np
import pytest
import torch

from skerch.distributed_measurements import DistributedHDF5, ssrft_idx_torch
from skerch.ssrft import SSRFT
from skerch.utils import gaussian_noise, torch_dtype_as_str

from . import rng_seeds


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def f64_atol():
    """Absolute error tolerance for float64."""
    result = {torch.float64: 1e-30}
    return result


@pytest.fixture
def f32_atol():
    """Absolute error tolerance for float32."""
    result = {torch.float32: 1e-5}
    return result


@pytest.fixture
def shapes_numfiles():
    """Shapes of individual HDF5 files and how many individual files."""
    result = [
        ((1,), 1),
        ((1,), 10),
        ((10,), 100),
        ((1000,), 100),
        #
        ((1, 2), 1),
        ((1, 2), 10),
        ((10, 2), 100),
        ((1000, 2), 100),
    ]
    return result


@pytest.fixture
def shapes_too_many_files(request):
    """Provides virtual datasets with too many files to cause errors."""
    result = [((5, 1), 20000)]
    if request.config.getoption("--skip_toomanyfiles"):
        result = []
    return result


@pytest.fixture
def shapes_measurements():
    """Shape of linear operator and SSRFT height.

    Since SSRFT can only be fat, SSRFT height must always be <= than
    the smallest dimension in shape.
    """
    result = [
        ((1, 1), 1),
        ((1, 10), 1),
        ((10, 1), 1),
        ((10, 10), 1),
        ((10, 10), 5),
        ((10, 10), 10),
        ((10, 100), 10),
        ((100, 100), 50),
        ((100, 1000), 50),
        ((1000, 1000), 500),
    ]
    return result


# ##############################################################################
# # DISTRIBUTED MEASUREMENTS
# ##############################################################################
def test_ssrft_idx_torch(rng_seeds, f64_atol, f32_atol, shapes_measurements):
    """Test case for applying a SSRFT indexed measurement.

    Creates a random matrix, and a SSRFT linop with given shapes. Tests that:

    * Applying SSRFT in indexed fashion is same as using @, both for left- and
      right-matmul
    """
    for seed in rng_seeds:
        ssrft_seed = seed + 1
        for dtype, atol in {**f32_atol, **f64_atol}.items():
            for lop_shape, num_measurements in shapes_measurements:
                lop_h, lop_w = lop_shape
                lop = gaussian_noise(
                    (lop_h, lop_w), seed=seed, dtype=dtype, device="cpu"
                )
                #
                for adjoint in (True, False):
                    ssrft_w = lop_h if adjoint else lop_w
                    ssrft = SSRFT((num_measurements, ssrft_w), seed=ssrft_seed)
                    #
                    target = (ssrft @ lop) if adjoint else (ssrft @ lop.T).T
                    # check that target shape is as expected
                    if adjoint:
                        expected_shape = (num_measurements, lop_w)
                    else:
                        expected_shape = (lop_h, num_measurements)
                    assert target.shape == expected_shape, "Wrong result shape!"
                    #
                    measurements = np.zeros_like(target)
                    flags = ["" for _ in range(num_measurements)]
                    flags = np.empty(num_measurements, dtype="str")
                    for i in range(num_measurements):
                        ssrft_idx_torch(
                            i,
                            num_measurements,
                            lop,
                            lop.device,
                            dtype,
                            measurements[i] if adjoint else measurements[:, i],
                            seed=ssrft_seed,
                            flag=flags[i : i + 1],
                            processing_flag="0",
                            success_flag="1",
                            adjoint=adjoint,
                        )
                    assert np.allclose(
                        measurements, target, atol=atol
                    ), "distributed SSRFT produced wrong result!"
                    assert all(
                        (f == "1" for f in flags)
                    ), "distributed SSRFT produced unsuccessful flags!"


# ##############################################################################
# # DISTRIBUTED HDF5
# ##############################################################################
def test_distributed_files(  # noqa: C901  # ignore "is too complex"
    rng_seeds, f64_atol, f32_atol, shapes_numfiles
):
    """Test case for ``DistributedHDF5`` creation and write/read.

    Creates a temporary DistributedHDF5 with random values, and checks that:
    * Files are actually created
    * Values can be written
    * Shapes of sub-datasets and main dataset are as expected
    * All values between sub-datasets, main dataset and test are consistent
    """
    for seed in rng_seeds:
        for dtype, atol in {**f32_atol, **f64_atol}.items():
            for shape, num_files in shapes_numfiles:
                for filedim_last in (True, False):
                    # target tensor
                    target_vals = torch.empty((num_files, *shape), dtype=dtype)
                    for i in range(num_files):
                        target_vals[i] = gaussian_noise(
                            shape, seed=seed + i, dtype=dtype, device="cpu"
                        )

                    if filedim_last:
                        target_vals = target_vals.moveaxis(0, -1)
                    # check dataset is actually being created on (temp) disk
                    tmpdir = tempfile.TemporaryDirectory()
                    out_path = os.path.join(tmpdir.name, "distdata_{}.h5")
                    h5_path, h5_subpaths = DistributedHDF5.create(
                        out_path,
                        num_files,
                        shape,
                        torch_dtype_as_str(dtype),
                        filedim_last=filedim_last,
                    )
                    assert os.path.isfile(h5_path), "Merged HDF5 not created?"
                    for sp in h5_subpaths:
                        assert os.path.isfile(sp), "Partial HDF5 not created?"
                    # write values to sub-datasets
                    for i in range(num_files):
                        vals, flag, h5 = DistributedHDF5.load(h5_subpaths[i])
                        vals[:] = gaussian_noise(
                            shape, seed=seed + i, dtype=dtype, device="cpu"
                        )
                        flag[0] = str(i)
                        h5.close()
                    # test that separate files and merger are correct
                    all_vals, all_flags, all_h5 = DistributedHDF5.load_virtual(
                        h5_path
                    )
                    assert (
                        all_vals.shape == target_vals.shape
                    ), "Wrong merged dataset shape!"
                    assert np.allclose(
                        target_vals, all_vals, atol=atol
                    ), "Wrong merged dataset values!"
                    assert all_flags.shape == (
                        num_files,
                    ), "Wrong merged flags shape!"
                    assert all_flags
                    for i in range(num_files):
                        vals, flag, h5 = DistributedHDF5.load(h5_subpaths[i])
                        assert vals.shape == shape, "Wrong sub-dataset shape!"
                        int_flag = int(flag[0].decode())
                        assert i == int_flag, "Wrong flag value!"
                        #
                        if filedim_last:
                            sub_target = target_vals[..., i]
                            sub_all = all_vals[..., i]
                        else:
                            sub_target = target_vals[i]
                            sub_all = all_vals[i]
                        assert np.allclose(
                            vals, sub_target, atol=atol
                        ), "Sub-dataset inconsistent with target!"
                        assert np.allclose(
                            vals, sub_all, atol=atol
                        ), "Sub-dataset inconsistent with merged dataset!"
                        h5.close()
                    #
                    all_h5.close()
                    tmpdir.cleanup()


def test_too_many_files(shapes_too_many_files):
    """Test case for too large ``DistributedHDF5``: read errors.

    Creates a temporary DistributedHDF5 with random values with a lot of
    files, and checks that, while all other tests still pass, loading the
    full virtual dataset actually contains erroneous (missing) columns.
    """
    seed = 0
    dtype, atol = torch.float32, 1e-5
    filedim_last = False
    for shape, num_files in shapes_too_many_files:
        # target tensor
        target_vals = torch.empty((num_files, *shape), dtype=dtype)
        for i in range(num_files):
            target_vals[i] = gaussian_noise(
                shape, seed=seed + i, dtype=dtype, device="cpu"
            )
        if filedim_last:
            target_vals = target_vals.moveaxis(0, -1)
        # check dataset is actually being created on (temp) disk
        tmpdir = tempfile.TemporaryDirectory()
        out_path = os.path.join(tmpdir.name, "distdata_{}.h5")
        h5_path, h5_subpaths = DistributedHDF5.create(
            out_path,
            num_files,
            shape,
            torch_dtype_as_str(dtype),
            filedim_last=filedim_last,
        )
        assert os.path.isfile(h5_path), "Merged HDF5 not created?"
        for sp in h5_subpaths:
            assert os.path.isfile(sp), "Partial HDF5 not created?"
        # write values to sub-datasets
        for i in range(num_files):
            vals, flag, h5 = DistributedHDF5.load(h5_subpaths[i])
            vals[:] = gaussian_noise(
                shape, seed=seed + i, dtype=dtype, device="cpu"
            )
            flag[0] = str(i)
            h5.close()
        # test that separate files  are correct
        all_vals, all_flags, all_h5 = DistributedHDF5.load_virtual(h5_path)
        assert (
            all_vals.shape == target_vals.shape
        ), "Wrong merged dataset shape!"
        # but here is the catch, merger is not correct!
        # this is due to file limit being exceeded, overflown values are empty
        with pytest.raises(AssertionError):
            assert np.allclose(
                target_vals, all_vals, atol=atol
            ), "Wrong merged dataset values!"
        # it would follow to test that merged is not consistent with
        # sub-datasets, but this itself causes OSErros, so we skip that.
        all_h5.close()
        tmpdir.cleanup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""CLI plugin to merge distributed HDF5 measurements into a single one."""


from ..distributed_measurements import DistributedHDF5


# ##############################################################################
# # ENTRY POINT
# ##############################################################################
def main(
    virt_path,
    out_path=None,
    ok_flag=None,
    delete_subfiles=True,
):
    """See :meth:`.distributed_measurements.DistributedHDF5.merge_all`."""
    # merge virtual HDF5 datasets into monolithic without memory overhead
    merged_path = DistributedHDF5.merge_all(
        virt_path,
        out_path=out_path,
        check_success_flag=ok_flag,
        delete_subfiles_while_merging=delete_subfiles,
    )
    print(merged_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Utilities for synthetic (random) matrices.

This module implements a few families of synthetic random matrices, inspired by
`[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_, but also expanded:

* Low rank + noise: dampened Gaussian noise matrix in the form ``G @ G.T``,
  plus a (low-rank) subset of the identity matrix.
* Exponential: Singular values have a low-rank unit section followed by
  exponential decay.
* Polynomial: Singular values have a low-rank unit section followed by
  polynomial decay.

In `[TYUC2019] <https://arxiv.org/abs/1902.08651>`_, the exponential and
polynomial matrices are diagonal. here, we extend them as follows:

* Random orthogonal matrices for the left and right singular spaces
* Optionally, the matrix is square and symmetric so left and right vectors are
  adjoint
* Optionally, the decaying spectrum is multiplied by Rademacher noise, to test
  also with non-PSD matrices
"""


import torch

from .utils import gaussian_noise, rademacher_flip


# ##############################################################################
# # SYNTH MATRIX FACTORY
# ##############################################################################
class SynthMat:
    """Static class to produce different families of synthetic matrices."""

    @staticmethod
    def lowrank_noise(
        shape=(100, 100),
        rank=10,
        snr=1e-4,
        seed=0b1110101001010101011,
        dtype=torch.float64,
        device="cpu",
    ):
        """Low rank + noise random matrix.

        Produces a lowrank+noise matrix, as follows: First, A Gaussian IID noise
        matrix ``G`` is sampled for the provided (square) shape. Then, the
        noisy matrix ``(snr / dims) * G@G.T`` is computed. Finally, the first
        ``rank`` diagonal entries are incremented by 1.
        As a result, we have a symmetric noisy matrix with strong diagonal
        for the first ``rank`` entries, and with a signal-to-noise ratio of
        ``snr``.

        :param int rank: How many diagonal entries should be reinforced.
        :param float snr: Signal-to-noise ratio. In
          `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_ the following
          values are used: 1e-4 for low noise, 1e-2 for mid noise, and 1e-1 for
          high noise.
        :returns: A matrix of given shape and properties.
        """
        h, w = shape
        assert h == w, "lowrank_noise must be square! (and symmetric)"
        # create matrix as a scaled outer product of Gaussian noise
        result = gaussian_noise(
            shape,
            mean=0.0,
            std=1.0,
            seed=seed,
            dtype=dtype,
            device=device,
        )
        result = (snr / shape[0]) * (result @ result.T)
        # add 1 to the first "rank" diagonal entries
        result[range(rank), range(rank)] += 1
        return result

    @staticmethod
    def _decay_helper(
        svals,
        shape=(100, 100),
        rank=10,
        decay=0.5,
        symmetric=True,
        seed=0b1110101001010101011,
        dtype=torch.float64,
        device="cpu",
    ):
        """Given singular values, produces a random matrix.

        Helper method for the polynomial and exp-decay matrices. Given the
        singular values, it samples the left- and right- singular spaces,
        and returns the final matrix.
        """
        min_shape = min(shape)
        # build singular bases using QR subgroup algorithm (Diaconis). QR is not
        # fastest, but these are test matrices so speed is not crucial.
        G_left = gaussian_noise(
            (shape[0], min_shape),
            mean=0.0,
            std=1.0,
            seed=seed,
            dtype=dtype,
            device=device,
        )
        U, _ = torch.linalg.qr(G_left)
        del G_left
        #
        if symmetric:
            result = U @ torch.diag(svals) @ U.T
        else:
            G_right = gaussian_noise(
                (shape[1], min_shape),
                mean=0.0,
                std=1.0,
                seed=seed + 1,
                dtype=dtype,
                device=device,
            )
            V, _ = torch.linalg.qr(G_right)
            result = U @ torch.diag(svals) @ V.T
        #
        return result

    @classmethod
    def poly_decay(
        cls,
        shape=(100, 100),
        rank=10,
        decay=0.5,
        symmetric=True,
        seed=0b1110101001010101011,
        dtype=torch.float64,
        device="cpu",
        psd=True,
    ):
        r"""Random matrix with polynomial singular value decay.

        Produces a matrix in the form ``U @ S @ V.T``, where ``U`` and ``V`` are
        random orthogonal matrices, and ``S`` has entries with polynomially
        decaying magnitude, analogous to the ones described in
        `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_.

        :param int rank: Entries in ``S[:rank]`` will have a magnitude of 1.
        :param float decay: Parameter determining how quickly magnitudes in
          ``S[rank:]`` decay to zero, following :math:`2^d, 3^d, 4^d, \dots`
          for decay :math:`d`. In
          `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_ the following
          values are used: 0.5 for slow decay, 1 for medium, 2 for fast.
        :param bool symmetric: If true, ``V == U``.
        :param psd: If false, and matrix is symmetric, the "tail" singular
          values will be multiplied with Rademacher noise to create a non-PSD
          matrix.
        """
        min_shape = min(shape)
        # a few ones, followed by a poly decay
        svals = torch.zeros(min_shape, dtype=dtype).to(device)
        svals[:rank] = 1
        svals[rank:] = torch.arange(2, min_shape - rank + 2) ** (-float(decay))
        if (not psd) and symmetric:
            rademacher_flip(svals[rank:], seed=seed + 1, inplace=True)
        #
        result = cls._decay_helper(
            svals, shape, rank, decay, symmetric, seed, dtype, device
        )
        return result

    @classmethod
    def exp_decay(
        cls,
        shape=(100, 100),
        rank=10,
        decay=0.5,
        symmetric=True,
        seed=0b1110101001010101011,
        dtype=torch.float64,
        device="cpu",
        psd=True,
    ):
        r"""Random matrix with exponential singular value decay.

        Produces a matrix in the form ``U @ S @ V``, where ``U`` and ``V`` are
        random orthogonal matrices, and ``S`` has entries with exponentially
        decaying magnitude, analogous to the ones described in
        `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_.

        :param int rank: Entries in ``S[:rank]`` will have a magnitude of 1.
        :param float decay: Parameter determining how quickly magnitudes in
          ``S[rank:]`` decay to zero, following :math:`10^{-d}, 10^{-2d}, \dots`
          for decay :math:`d`. In
          `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_ the following
          values are used: 0.01 for slow decay, 0.1 for medium, 0.5 for fast.
        :param bool symmetric: If true, ``V == U``.
        :param psd: If false, and matrix is symmetric, the "tail" singular
          values will be multiplied with Rademacher noise to create a non-PSD
          matrix.
        """
        min_shape = min(shape)
        # a few ones, followed by exp decay
        svals = torch.zeros(min_shape, dtype=dtype).to(device)
        svals[:rank] = 1
        svals[rank:] = 10 ** -(decay * torch.arange(1, min_shape - rank + 1))
        if (not psd) and symmetric:
            rademacher_flip(svals[rank:], seed=seed + 1, inplace=True)
        #
        result = cls._decay_helper(
            svals, shape, rank, decay, symmetric, seed, dtype, device
        )
        return result



# -*- coding: utf-8 -*-

r"""Sketched Triangular Operators
=================================

In this tutorial we create a noisy matrix and perform sketched estimations of
matrix-vector multiplications with its lower and upper triangles.
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import torch

from skerch.synthmat import SynthMat
from skerch.triangles import TriangularLinOp
from skerch.utils import gaussian_noise

# %%
#
# ##############################################################################
#
# Globals and creation of test matrix
# -----------------------------------
#
# We will create a random matrix of shape ``(ORDER, ORDER)``, and a test vector
# with Gaussian i.i.d. noise (the other global parameters can be ignored).

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
DTYPE = torch.float64
ORDER, RANK = 1000, 100
EXP_DECAY = 0.2
SEED = 12345

v = gaussian_noise(ORDER, seed=SEED - 1, dtype=DTYPE, device=DEVICE)

linop = SynthMat.exp_decay(
    shape=(ORDER, ORDER),
    rank=RANK,
    decay=EXP_DECAY,
    symmetric=False,
    seed=SEED,
    dtype=DTYPE,
    device=DEVICE,
    psd=False,
)

tril = torch.tril(linop, diagonal=0)
triu = torch.triu(linop, diagonal=0)

# test measurements
lo_v = tril @ v
v_lo = v @ tril
up_v = triu @ v
v_up = v @ triu


# %%
#
# ##############################################################################
#
# Sketched triangular operators
# -----------------------------
#
# To estimate triangular matrix-vector products, these linear operators perform
# multiple measurements (as opposed to just one if we knew the exact triangular
# operator). These measurements have two very distinct natures:
#
# * The "step-wise" measurements are exact, and measure blocks that are fully
#   inside of the triangle
# * The Hutchinson measurements are a modification of the Hutchinson diagonal
#   estimator, to estimate the product with the block-triangular patterns that
#   were ignored by the exact step-wise measurements.
#
# For this reason, we want to have as many exact measurements as we can afford.
# The Hutchinson measurements, much more inefficient, will be necessary if a
# relevant part of the linop is concentrated near the main diagonal.

tril_est = TriangularLinOp(
    linop,
    stair_width=(ORDER // 20),  # we do approx 20 exact measurements
    num_hutch_measurements=(ORDER // 5),  # Hutch works better for larger order
    lower=True,
    with_main_diagonal=True,
)

triu_est = TriangularLinOp(
    linop,
    stair_width=(ORDER // 20),
    num_hutch_measurements=(ORDER // 5),
    lower=False,
    with_main_diagonal=True,
)


lo_v_est = tril_est @ v
v_lo_est = v @ tril_est
up_v_est = triu_est @ v
v_up_est = v @ triu_est


fig, (ax1, ax2) = plt.subplots(ncols=2)
ax1.imshow(
    triu.cpu()[:50, :50],
    cmap="bwr",
    aspect="auto",
    norm=mpl.colors.CenteredNorm(),
)
ax2.plot(up_v.cpu()[:50], color="black", label="original")
ax2.plot(up_v_est.cpu()[:50], color="pink", label="approximation")
ax2.legend()
fig.suptitle("Upper Triangular and $triu(A)~v$ (detail)")
fig.tight_layout()
#

fig, (ax1, ax2) = plt.subplots(ncols=2)
ax1.imshow(
    tril.cpu()[:, :], cmap="bwr", aspect="auto", norm=mpl.colors.CenteredNorm()
)
ax2.plot(lo_v.cpu()[:], color="black", label="original")
ax2.plot(lo_v_est.cpu()[:], color="pink", label="approximation")
ax2.legend()
fig.suptitle("Lower Triangular and $tril(A)~v$")
fig.tight_layout()
#
fig, (ax1, ax2) = plt.subplots(ncols=2)
ax1.imshow(
    tril.cpu()[:, :], cmap="bwr", aspect="auto", norm=mpl.colors.CenteredNorm()
)
ax2.plot(v_lo.cpu()[:], color="black", label="original")
ax2.plot(v_lo_est.cpu()[:], color="pink", label="approximation")
ax2.legend()
fig.suptitle("Lower Triangular and $v~tril(A)$")
fig.tight_layout()
#
fig, (ax1, ax2) = plt.subplots(ncols=2)
ax1.imshow(
    triu.cpu()[:, :], cmap="bwr", aspect="auto", norm=mpl.colors.CenteredNorm()
)
ax2.plot(up_v.cpu()[:], color="black", label="original")
ax2.plot(up_v_est.cpu()[:], color="pink", label="approximation")
ax2.legend()
fig.suptitle("Upper Triangular and $triu(A)~v$")
fig.tight_layout()
#
fig, (ax1, ax2) = plt.subplots(ncols=2)
ax1.imshow(
    triu.cpu()[:, :], cmap="bwr", aspect="auto", norm=mpl.colors.CenteredNorm()
)
ax2.plot(v_up.cpu()[:], color="black", label="original")
ax2.plot(v_up_est.cpu()[:], color="pink", label="approximation")
ax2.legend()
fig.suptitle("Upper Triangular and $v~triu(A)$")
fig.tight_layout()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""A-priori sketch hyperparameters.

CLI plugin to provide optimal number of measurements for a given budget and
linear operator.
"""

from ..a_priori import a_priori_hyperparams


# ##############################################################################
# # ENTRY POINT
# ##############################################################################
def main(shape, budget, hermitian=False):
    """See :func:`.a_priori.a_priori_hyperparams`."""
    outer, inner = a_priori_hyperparams(
        shape, budget, complex_data=False, hermitian=hermitian
    )
    print("Shape:", shape)
    print("Budget:", budget)
    print("Hermitian:", hermitian)
    print("  Outer measurements:", outer)
    print("  Inner measurements:", inner)
    print(
        f"[CHECK]: {outer} * {sum(shape)} + {inner}**2 =",
        (outer * sum(shape)) + inner**2,
    )
    print(
        f"[CHECK]: {outer + 1} * {sum(shape)} + {inner}**2 =",
        ((outer + 1) * sum(shape)) + inner**2,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""General utilities for the ``skerch`` library."""


import torch


# ##############################################################################
# # STR
# ##############################################################################
def torch_dtype_as_str(dtype):
    """Torch dtype to string.

    Given a PyTorch datatype object, like ``torch.float32``, returns the
    corresponding string, in this case 'float32'.
    """
    full_str = str(dtype)
    dot_idx = full_str.index(".")
    result = full_str[dot_idx + 1 :]
    return result


# ##############################################################################
# # ERRORS
# ##############################################################################
class NoFlatError(Exception):
    """Error to be thrown when a tensor is not flat (i.e. not a vector)."""

    pass


class BadShapeError(Exception):
    """Error to be thrown when a shape is not as it should."""

    pass


# ##############################################################################
# # REPRODUCIBLE NOISE
# ##############################################################################
def uniform_noise(shape, seed=None, dtype=torch.float64, device="cpu"):
    """Reproducible ``torch.rand`` uniform noise.

    :returns: A tensor of given shape, dtype and device, containing uniform
      random noise between 0 and 1 (analogous to ``torch.rand``), but with
      reproducible behaviour fixed to given random seed.
    """
    rng = torch.Generator(device=device)
    rng.manual_seed(seed)
    noise = torch.rand(shape, generator=rng, dtype=dtype, device=device)
    return noise


def gaussian_noise(
    shape, mean=0.0, std=1.0, seed=None, dtype=torch.float64, device="cpu"
):
    """Reproducible ``torch.normal`` Gaussian noise.

    :returns: A tensor of given shape, dtype and device, containing gaussian
      noise with given mean and std (analogous to ``torch.normal``), but with
      reproducible behaviour fixed to given random seed.
    """
    rng = torch.Generator(device=device)
    rng.manual_seed(seed)
    #
    noise = torch.zeros(shape, dtype=dtype, device=device)
    noise.normal_(mean=mean, std=std, generator=rng)
    return noise


def rademacher_noise(shape, seed=None, device="cpu"):
    """Reproducible Rademacher noise.

    .. note::
      This function makes use of :func:`uniform_noise` to sample the Rademacher
      noise. If ``x`` itself has been generated using ``uniform_noise``, make
      sure to use a different seed to mitigate correlations.

    .. warning::
      PyTorch does not ensure RNG reproducibility across
      devices. This parameter determines the device to generate the noise from.
      If you want cross-device reproducibility, make sure that the noise is
      always generated from the same device.

    :param shape: Shape of the output tensor with Rademacher noise.
    :param seed: Seed for the randomness.
    :param device: Device of the output tensor and also source for the noise.
      See warning.
    """
    noise = (
        uniform_noise(shape, seed=seed, dtype=torch.float32, device=device)
        > 0.5
    ) * 2 - 1
    return noise


def randperm(n, seed=None, device="cpu", inverse=False):
    """Reproducible randperm of ``n`` integers from  0 to (n-1) (both included).

    :param bool inverse: If False, a random permutation ``P`` is provided. If
      true, an inverse permutation ``Q`` is provided, such that both
      permutations are inverse to each other, i.e. ``v == v[P][Q] == v[Q][P]``.
    """
    rng = torch.Generator(device=device)
    rng.manual_seed(seed)
    #
    perm = torch.randperm(n, generator=rng, device=device)
    if inverse:
        # we take the O(N) approach since we anticipate large N
        inv = torch.empty_like(perm)
        inv[perm] = torch.arange(perm.size(0), device=perm.device)
        perm = inv
    return perm


def rademacher_flip(x, seed=None, inplace=True, rng_device="cpu"):
    """Reproducible random sign flip using Rademacher noise.

    .. note::
      This function makes use of :func:`uniform_noise` to sample the Rademacher
      noise. If ``x`` itself has been generated using ``uniform_noise``, make
      sure to use a different seed to mitigate correlations.

    .. warning::
      See :func:`rademacher_noise` for notes on reproducibility and more info.
    """
    mask = rademacher_noise(x.shape, seed, device=rng_device).to(x.device)
    if inplace:
        x *= mask
        return x, mask
    else:
        return x * mask, mask



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for linop (linear operator) utils."""


import pytest
import torch

from skerch.a_posteriori import a_posteriori_error
from skerch.linops import CompositeLinOp, DiagonalLinOp
from skerch.utils import BadShapeError, gaussian_noise

from . import rng_seeds, torch_devices


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def f64_atol():
    """Absolute error tolerance for float64."""
    result = {torch.float64: 1e-10}
    return result


@pytest.fixture
def f32_atol():
    """Absolute error tolerance for float32."""
    result = {torch.float32: 1e-5}
    return result


@pytest.fixture
def composite_sizes(request):
    """Chained shapes for the composite linear operator."""
    result = [
        (1, 1, 1),
        (1, 10, 1),
        (10, 1, 10),
        (1, 10, 100, 10, 1),
        (10, 1, 100, 1, 10),
        (1, 100, 1000, 1, 10),
        (100, 100, 100, 100, 100),
        (1000, 1000, 1000),
        (1000, 10, 1000),
        (10, 1000, 1000, 10),
    ]
    if request.config.getoption("--quick"):
        result = result[:5]
    return result


@pytest.fixture
def diagonal_sizes(request):
    """Dimensions for the diagonal linear operator."""
    result = [1, 3, 10, 30, 100, 300, 1000]
    if request.config.getoption("--quick"):
        result = result[:5]
    return result


@pytest.fixture
def num_test_measurements():
    """Number of a-posteriori measurements.

    Calling a_posteriori_error_bounds(100, 1) yields
    {'P(err<=0x)': 0.0, 'P(err>=2x)': 2.171579274145306e-07}
    So if we obtain a small number, chances that it is not actually small are
    extremely low.
    """
    return 100


# ##############################################################################
# # TESTS
# ##############################################################################
def test_composite_correctness(
    rng_seeds,
    torch_devices,
    f64_atol,
    f32_atol,
    composite_sizes,
    num_test_measurements,
):
    """Test case for correctness of ``CompositeLinOp``.

    Creates a chain of random submatrices, and then composes them explicity
    using ``CompositeLinOp``. Tests that:

    * Composed shapes are as expected
    * Explicit and implicit compositions are identical both with left and right
      matrix multiplication
    """
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype, atol in {**f32_atol, **f64_atol}.items():
                for sizes in composite_sizes:
                    # sample random submatrices
                    submatrices = []
                    for i, (h, w) in enumerate(zip(sizes[:-1], sizes[1:])):
                        mat = gaussian_noise(
                            (h, w), seed=seed + i, dtype=dtype, device=device
                        )
                        submatrices.append((f"{(h, w)}", mat))
                    # compose submatrices, either explicit or implicit
                    explicit = submatrices[0][1]
                    for _, mat in submatrices[1:]:
                        explicit = explicit @ mat
                    composite = CompositeLinOp(submatrices)
                    # check that shapes are correct
                    expected_shape = (sizes[0], sizes[-1])
                    assert (
                        composite.shape == expected_shape
                    ), "Incorrect composite shape?"
                    assert (
                        explicit.shape == expected_shape
                    ), "Incorrect explicit shape?"
                    # check that explicit and implicit are identical for both
                    # left and right matmul
                    for adjoint in [True, False]:
                        (_, _, err_estimate), _ = a_posteriori_error(
                            explicit,
                            composite,
                            num_test_measurements,
                            seed=seed + len(sizes),
                            dtype=dtype,
                            device=device,
                            adjoint=adjoint,
                        )
                        assert (
                            err_estimate / explicit.numel()
                        ) <= atol, "Incorrect composite operation!"


def test_composite_shapes(torch_devices, f32_atol, composite_sizes):
    """Test case for shape consistency of ``CompositeLinOp``.

    Checks that:
    * Mismatching matrix shapes give bad shape error
    * Bad input shapes (left and right) give bad shape error
    * Correct inputs can be both vectors and matrices
    """
    for device in torch_devices:
        for dtype, _atol in f32_atol.items():
            for sizes in composite_sizes:
                # create composite matrix with mismatching shapes, should
                # throw an error
                submatrices = []
                for i, (h, w) in enumerate(zip(sizes[:-1], sizes[1:])):
                    mat = gaussian_noise(
                        (h, w + 1), seed=i, dtype=dtype, device=device
                    )
                    submatrices.append((f"{(h, w)}", mat))
                with pytest.raises(BadShapeError):
                    implicit = CompositeLinOp(submatrices)
                # now create composite matrix with correct shapes, but feed
                # vectors with incorrect shapes. Should also throw an error
                submatrices = []
                for i, (h, w) in enumerate(zip(sizes[:-1], sizes[1:])):
                    mat = gaussian_noise(
                        (h, w), seed=i, dtype=dtype, device=device
                    )
                    submatrices.append((f"{(h, w)}", mat))
                implicit = CompositeLinOp(submatrices)
                comp_h, comp_w = implicit.shape
                #
                x = gaussian_noise(
                    comp_h + 1,
                    seed=len(sizes) + 1,
                    dtype=dtype,
                    device=device,
                )
                with pytest.raises(RuntimeError):
                    x @ implicit
                #
                x = gaussian_noise(
                    comp_w + 1,
                    seed=len(sizes) + 1,
                    dtype=dtype,
                    device=device,
                )
                with pytest.raises(RuntimeError):
                    implicit @ x
                # correct matmul shape against vectors
                x = gaussian_noise(comp_w, seed=0, dtype=dtype, device=device)
                assert (implicit @ x).shape == (
                    comp_h,
                ), "Unexpected right matmul shape!"
                x = gaussian_noise(comp_h, seed=0, dtype=dtype, device=device)
                assert (x @ implicit).shape == (
                    comp_w,
                ), "Unexpected left matmul shape!"
                # correct matmul shape against matrices
                x = gaussian_noise(
                    (comp_w, 2), seed=0, dtype=dtype, device=device
                )
                assert (implicit @ x).shape == (
                    comp_h,
                    2,
                ), "Unexpected right matmul shape!"

                x = gaussian_noise(
                    (2, comp_h), seed=0, dtype=dtype, device=device
                )
                assert (x @ implicit).shape == (
                    2,
                    comp_w,
                ), "Unexpected left matmul shape!"


def test_diagonal_correctness(
    rng_seeds,
    torch_devices,
    f64_atol,
    f32_atol,
    diagonal_sizes,
    num_test_measurements,
):
    """Test case for correctness of ``DiagonalLinOp``.

    Create a random vector, and build a diagonal matrix either explicitly or
    using ``DiagonalLinOp``. Test that:

    * Shapes are as expected
    * Explicit and implicit matrices are identical both with left and right
      matrix multiplication
    """
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype, atol in {**f32_atol, **f64_atol}.items():
                for size in diagonal_sizes:
                    # sample random diagonal
                    diag = gaussian_noise(
                        size, seed=seed, dtype=dtype, device=device
                    )
                    explicit = torch.diag(diag)
                    implicit = DiagonalLinOp(diag)
                    # check that shapes are correct
                    expected_shape = (size, size)
                    assert (
                        implicit.shape == expected_shape
                    ), "Incorrect implicit shape?"
                    assert (
                        explicit.shape == expected_shape
                    ), "Incorrect explicit shape?"
                    # check that explicit and implicit are identical for both
                    # left and right matmul
                    for adjoint in [True, False]:
                        (_, _, err_estimate), _ = a_posteriori_error(
                            explicit,
                            implicit,
                            num_test_measurements,
                            seed=seed + 1,
                            dtype=dtype,
                            device=device,
                            adjoint=adjoint,
                        )
                        assert (
                            err_estimate / explicit.numel()
                        ) <= atol, "Incorrect composite operation!"


def test_diagonal_shapes(torch_devices, f32_atol, diagonal_sizes):
    """Test case for shape consistencies of ``DiagonalLinOp``.

    Check that
    * Bad input shapes (left and right) give bad shape error
    * Correct inputs can be both vectors and matrices
    """
    for device in torch_devices:
        for dtype, _atol in f32_atol.items():
            for size in diagonal_sizes:
                # sample random diagonal
                diag = gaussian_noise(size, seed=1, dtype=dtype, device=device)
                implicit = DiagonalLinOp(diag)
                x = gaussian_noise(size + 1, seed=2, dtype=dtype, device=device)
                # if the commands below throw a BadShapeError, the test passes
                with pytest.raises(BadShapeError):
                    implicit @ x
                with pytest.raises(BadShapeError):
                    x @ implicit
                # correct matmul shape against vectors
                x = gaussian_noise(size, seed=2, dtype=dtype, device=device)
                assert (implicit @ x).shape == (
                    size,
                ), "Unexpected right matmul shape!"
                assert (x @ implicit).shape == (
                    size,
                ), "Unexpected left matmul shape!"
                # correct matmul shape against matrices
                x = gaussian_noise(
                    (size, 2), seed=2, dtype=dtype, device=device
                )
                assert (implicit @ x).shape == (
                    size,
                    2,
                ), "Unexpected right matmul shape!"
                assert (x.T @ implicit).shape == (
                    2,
                    size,
                ), "Unexpected left matmul shape!"



# -*- coding: utf-8 -*-

r"""Deep Learning
=================

As shown in e.g. :ref:`Sketched Decompositions`, this library computes sketched
decompositions with any linear operator that implements left- and right matrix
multiplication via the ``@`` operator, and the ``shape = (height, width)``
attribute.

The amazing `CurvLinOps library <https://github.com/f-dangel/curvlinops>`_
provides linear operators that satisfy this property (among many others), for a
variety of very useful DL-related linear operators, like the Hessian, the
Jacobian and the Generalized Gauss-Newton (GGN). It is also implemented
with PyTorch as a backend, but the ``CurvLinOps`` operators actually implement
the ``SciPy LinearOperator`` interface, and can be therefore used with most of
the LinAlg routines available in SciPy.

But, to the date, SciPy does not provide sketched eigendecompositions for
such operators (see `here <https://github.com/scipy/scipy/issues/16049>`_) and
classical algorithms have potentially much worse runtime and memory
requirements.

In this tutorial, we show how to approximate a GPU-accelerated Hessian
full eigendecomposition from a deep neural network, using ``skerch``'s very own
SEIGH together with CurvLinOps. Note that this procedure can easily scale to
much larger networks (with parameters up to the millions) and datasets (with
thousands of samples) than traditional methods.

We also show via *a posteriori* methods that the recovered eigendecomposition
is pretty close to the original, **full** Hessian. No more compromises!
"""

import matplotlib.pyplot as plt
import torch
from curvlinops import HessianLinearOperator

from skerch.a_posteriori import a_posteriori_error
from skerch.decompositions import seigh, truncate_core
from skerch.linops import CompositeLinOp, DiagonalLinOp, TorchLinOpWrapper

# %%
#
# ##############################################################################
#
# Setup
# -----
#
# Mirroring CurvLinOps tutorials, we will use synthetic data, consisting of two
# mini-batches, a small MLP, and mean-squared error as loss function. But we
# will use a larger scale :)

# GLOBALS
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
# dl globals
N = 20
D_in = 50
D_hidden = 100
D_out = 10
# skerch globals
OUTER, INNER, TRUNCATE = 500, 1000, 400
NUM_A_POSTERIORI = 20


# synthetic dataset
X1, y1 = torch.rand(N, D_in).to(DEVICE), torch.rand(N, D_out).to(DEVICE)
X2, y2 = torch.rand(N, D_in).to(DEVICE), torch.rand(N, D_out).to(DEVICE)
data = [(X1, y1), (X2, y2)]

# neural network
model = torch.nn.Sequential(
    torch.nn.Linear(D_in, D_hidden),
    torch.nn.ReLU(),
    torch.nn.Linear(D_hidden, D_hidden),
    torch.nn.Sigmoid(),
    torch.nn.Linear(D_hidden, D_out),
).to(DEVICE)
params = [p for p in model.parameters() if p.requires_grad]
loss_function = torch.nn.MSELoss(reduction="mean").to(DEVICE)

# %%
#
# ##############################################################################
#
# Sketched Hessian Eigendecomposition
# -----------------------------------
#
# While ``CurvLinOps`` interfaces with ``SciPy/NumPy``, ``skerch`` does expect
# that the linear operators input and output ``PyTorch`` tensors. To overcome
# this discrepancy, ``skerch`` provides a
# :class:`skerch.linops.TorchLinOpWrapper` class. With this, ``seigh``
# can be directly applied to the wrapped ``CurvLinOps`` operator!


class TorchHessianLinearOperator(TorchLinOpWrapper, HessianLinearOperator):
    pass


# Create the Hessian LinOp, which is a function of model, loss and data
H = TorchHessianLinearOperator(model, loss_function, params, data)
print(H)

# Perform sketched Hermitian eigendecomposition!
Q, U, S = seigh(
    H,
    op_device=DEVICE,
    op_dtype=getattr(torch, str(H.dtype)),
    outer_dim=OUTER,
    inner_dim=INNER,
)

# Compose low-rank approximation
U, S = truncate_core(TRUNCATE, U, S)
appr_H = CompositeLinOp(
    (
        ("Q", Q),
        ("U", U),
        ("S", DiagonalLinOp(S)),
        ("Ut", U.T),
        ("Qt", Q.T),
    )
)


# Plot a fragment of the recovered Hessian
frag_n = 50
appr_H_fragment = torch.empty_like(Q[:frag_n, :frag_n])
buff = torch.zeros_like(Q[:, 0])
for i in range(frag_n):
    buff *= 0
    buff[-(i + 1)] = 1
    appr_H_fragment[:, -(i + 1)] = (appr_H @ buff)[-frag_n:]

plt.imshow(appr_H_fragment.cpu().numpy())
plt.show()

# %%
#
# ##############################################################################
#
# Error estimation
# ----------------
#
# Now all is left is to compute an *a posteriori* estimate of the Frobenius
# error between the original Hessian :math:`H` and our sketched, low-rank
# approximation :math:`\hat{H}`. The ``RELATIVE ERROR`` :math:`\rho` is given
# as the ratio
# :math:`\rho := \frac{\lVert H - \hat{H} \rVert_F}{\lVert H \rVert_F}`.

(f1, f2, frob_err) = a_posteriori_error(
    H,
    appr_H,
    NUM_A_POSTERIORI,
    dtype=getattr(torch, str(H.dtype)),
    device=DEVICE,
)[0]

print("Estimated Hessian Norm:", f1**0.5)
print("Estimated Frobenius Error:", frob_err**0.5)
print("RELATIVE ERROR:", (frob_err / f1) ** 0.5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Utilities for linear operators.

This PyTorch library is intended to work with (potentially matrix-free) linear
operators that support left- and right- matrix multiplication (via e.g.
``v @ lop`` and ``lop @ v``) as well as the ``.shape`` attribute.

This module implements basic support for said functionality, as well as some
default linear operators (composite, diagonal...).
"""


import torch

from .utils import BadShapeError, NoFlatError


# ##############################################################################
# # BASE LINEAR OPERATOR
# ##############################################################################
class BaseLinOp:
    """Base class for linear operators.

    Provides the ``.shape`` attribute and basic matmul functionality with
    vectors and matrices (also via the ``@`` operator). Intended to be
    extended with further functionality.
    """

    def __init__(self, shape):
        """:param shape: ``(height, width)`` of linear operator."""
        if len(shape) != 2:
            raise BadShapeError("Shape must be a (height, width) pair!")
        self.shape = shape

    def __repr__(self):
        """Returns a string in the form <classname(shape)>."""
        clsname = self.__class__.__name__
        s = f"<{clsname}({self.shape[0]}x{self.shape[1]})>"
        return s

    def check_input(self, x, adjoint):
        """Checking that input has compatible shape.

        :param x: The input to this linear operator.
        :param bool adjoint: If true, ``x @ self`` is assumed, otherwise
          ``self @ x``.
        """
        try:
            assert len(x.shape) in {
                1,
                2,
            }, "Only vector or matrix input supported"
            #
            if adjoint:
                assert (
                    x.shape[-1] == self.shape[0]
                ), f"Mismatching shapes! {x.shape} <--> {self.shape}"
            else:
                assert (
                    x.shape[0] == self.shape[1]
                ), f"Mismatching shapes! {self.shape} <--> {x.shape}"
        except AssertionError as ae:
            raise BadShapeError from ae

    def matmul(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        :param x: Expected a tensor of shape ``(w,)``.
          Note that shapes ``(w, k)`` will be automatically passed as ``k``
          vectors of length ``w``.
        """
        raise NotImplementedError

    def rmatmul(self, x):
        """Adjoint (left) matrix-vector multiplication ``x @ self``.

        :param x: Expected a tensor of shape ``(h,)``.
          Note that shapes ``(h, k)`` will be automatically passed as ``k``
          vectors of length ``h``.
        """
        raise NotImplementedError

    # operator interfaces
    def __matmul__(self, x):
        """Convenience wrapper to :meth:`.matmul`."""
        self.check_input(x, adjoint=False)
        try:
            return self.matmul(x)
        except NoFlatError:
            result = torch.zeros((self.shape[0], x.shape[1]), dtype=x.dtype).to(
                x.device
            )
            for i in range(x.shape[1]):
                result[:, i] = self.matmul(x[:, i])
            return result

    def __rmatmul__(self, x):
        """Convenience wrapper to :meth:`.rmatmul`."""
        self.check_input(x, adjoint=True)
        try:
            return self.rmatmul(x)
        except NoFlatError:
            result = torch.zeros((x.shape[0], self.shape[1]), dtype=x.dtype).to(
                x.device
            )
            for i in range(x.shape[0]):
                result[i, :] = self.rmatmul(x[i, :])
            return result

    def __imatmul__(self, x):
        """Assignment matmul operator ``@=``.

        .. note::
          This method is deactivated by default since linear operators may be
          matrix-free.
        """
        raise NotImplementedError("Matmul assignment not supported!")


# ##############################################################################
# # BASE LINEAR OPERATOR INVOLVING RANDOMNESS
# ##############################################################################
class BaseRandomLinOp(BaseLinOp):
    """Linear operators with pseudo-random behaviour.

    Like the base LinOp, but with a ``seed`` attribute that can be used to
    control random behaviour.
    """

    def __init__(self, shape, seed=0b1110101001010101011):
        """Instantiates a random linear operator.

        :param shape: ``(height, width)`` of linear operator.
        :param int seed: Seed for random behaviour.
        """
        super().__init__(shape)
        self.seed = seed

    def __repr__(self):
        """Returns a string in the form <classname(shape), seed>."""
        clsname = self.__class__.__name__
        s = f"<{clsname}({self.shape[0]}x{self.shape[1]}), seed={self.seed}>"
        return s


# ##############################################################################
# # COMPOSITE LINEAR OPERATOR
# ##############################################################################
class CompositeLinOp:
    """Composite linear operator.

    This class composes an ordered collection of operators ``[A, B, C, ...]``
    into ``A @ B @ C ...``.

    .. note::
      Using this class could be more inefficient than directly computing the
      composed operator, e.g. if ``A.shape = (1, 1000)`` and
      ``B.shape = (1000, 1)``, then computing the scalar ``C = A @ B`` and then
      applying it is more efficient than keeping a ``CompositeLinearoperator``
      with ``A, B`` (in terms of both memory and computation). This class does
      not check for such cases, users are encouraged to take this into account.
    """

    def __init__(self, named_operators):
        """Instantiates a composite linear operator.

        :param named_operators: Ordered collection in the form
          ``[(n_1, o_1), ...]`` where each ``n_i`` is a string with the name of
          operator ``o_i``. Each ``o_i`` operator must implement ``__matmul__``
          and ``__rmatmul__`` as well as the ``shape = (h, w)`` attribute. This
          object will then become the composite operator ``o_1 @ o_2 @ ...``
        """
        self.names, self.operators = zip(*named_operators)
        shapes = [o.shape for o in self.operators]
        for (_h1, w1), (h2, _w2) in zip(shapes[:-1], shapes[1:]):
            if w1 != h2:
                raise BadShapeError(f"Mismatching shapes in {shapes}!")
        self.shape = (shapes[0][0], shapes[-1][-1])

    def __matmul__(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        See parent class for more details.
        """
        result = self.operators[-1] @ x
        for o in reversed(self.operators[:-1]):
            result = o @ result
        return result

    def __rmatmul__(self, x):
        """Adjoint (left) matrix-vector multiplication ``x @ self``.

        See parent class for more details.
        """
        result = x @ self.operators[0]
        for o in self.operators[1:]:
            result = result @ o
        return result

    def __repr__(self):
        """Returns a string in the form op1 @ op2 @ op3 ..."""
        result = " @ ".join(self.names)
        return result


# ##############################################################################
# # DIAGONAL LINEAR OPERATOR
# ##############################################################################
class DiagonalLinOp(BaseLinOp):
    r"""Diagonal linear operator.

    Given a vector ``v`` of ``d`` dimensions, this class implements a diagonal
    linear operator of shape ``(d, d)`` via left- and right-matrix
    multiplication, as well as the ``shape`` attribute, only requiring linear
    (:math:`\mathcal{O}(d)`) memory and runtime.
    """

    MAX_PRINT_ENTRIES = 20

    def __init__(self, diag):
        """:param diag: Vector to be casted as diagonal linop."""
        if len(diag.shape) != 1:
            raise BadShapeError("Diag must be a vector!")
        self.diag = diag
        super().__init__((len(diag),) * 2)  # this sets self.shape also

    def __matmul__(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        See parent class for more details.
        """
        self.check_input(x, adjoint=False)
        if len(x.shape) == 2:
            result = (x.T * self.diag).T
        else:
            # due to torch warning, can't transpose shapes other than 2
            result = x * self.diag
        return result

    def __rmatmul__(self, x):
        """Adjoint (left) matrix-vector multiplication ``x @ self``.

        See parent class for more details.
        """
        self.check_input(x, adjoint=True)
        result = x * self.diag
        return result

    def __repr__(self):
        """Returns a string in the form <DiagonalLinOp(shape)[v1, v2, ...]>."""
        clsname = self.__class__.__name__
        diagstr = ", ".join(
            [str(x.item()) for x in self.diag[: self.MAX_PRINT_ENTRIES]]
        )
        if len(self.diag) > self.MAX_PRINT_ENTRIES:
            diagstr += "..."
        s = f"<{clsname}({self.shape[0]}x{self.shape[1]})[{diagstr}]>"
        return s


# ##############################################################################
# # DIAGONAL LINEAR OPERATOR
# ##############################################################################
class OrthProjLinOp(BaseLinOp):
    r"""Linear operator for an orthogonal projector.

    Given a "thin" matrix (i.e. height >= width) :math:`Q` of orthonormal
    columns, this class implements the orthogonal projector onto its span,
    i.e. :math:`Q Q^T`.
    """

    def __init__(self, Q):
        """Object initialization.

        :param Q: Linear operator of shape ``(h, w)`` with ``h >= w``, expected
          to be orthonormal (i.e. columns with unit Euclidean norm and all
          orthogonal to each other). It must implement the left and right
          matmul operations via the ``@`` operator.

        .. warning::

          This class assumes ``Q`` is orthonormal, but this is not checked.
        """
        self.Q = Q
        #
        h, w = Q.shape
        if w > h:
            raise BadShapeError("Projector matrix must be thin!")
        super().__init__((h, h))  # this sets self.shape also

    def __matmul__(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        See parent class for more details.
        """
        self.check_input(x, adjoint=False)
        result = self.Q @ (x.T @ self.Q).T
        return result

    def __rmatmul__(self, x):
        """Adjoint (left) matrix-vector multiplication ``x @ self``.

        See parent class for more details.
        """
        self.check_input(x, adjoint=True)
        result = self.Q @ (x @ self.Q).T
        return result.T

    def __repr__(self):
        """Returns a string in the form <OrthProjLinOp(Q_shape)>."""
        clsname = self.__class__.__name__
        s = f"<{clsname}({self.Q.shape})>"
        return s


class NegOrthProjLinOp(OrthProjLinOp):
    """Linear operator for a negative orthogonal projector.

    Given a "thin" matrix (i.e. height >= width) :math:`Q` of orthonormal
    columns, this class implements the orthogonal projector :math:`Q Q^T`.
    Given a "thin" matrix (i.e. height >= width) :math:`Q` of orthonormal
    columns, this class implements the orthogonal projector onto the space
    orthogonal to its span, i.e. :math:`(I - Q Q^T)`.
    """

    def __matmul__(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        See parent class for more details.
        """
        return x - super().__matmul__(x)

    def __rmatmul__(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        See parent class for more details.
        """
        return x - super().__rmatmul__(x)


# ##############################################################################
# # TORCH INTEROPERABILITY
# ##############################################################################
class TorchLinOpWrapper:
    """Linear operator that always accepts and produces PyTorch tensors.

    Since this library operates mainly with PyTorch tensors, but some useful
    LinOps interface with e.g. NumPy arrays instead, this mixin class acts as a
    wraper on the ``__matmul__`` and ``__rmatmul__`` operators,
    so that the operator expects and returns torch tensors, even when the
    wrapped operator interfaces with NumPy/HDF5. Usage example::

      # extend NumPy linear operator via multiple inheritance
      class TorchWrappedLinOp(TorchLinOpWrapper, LinOp):
          pass
      lop = TorchWrappedLinOp(...)  # instantiate normally
      w = lop @ v  # now v can be a PyTorch tensor
    """

    @staticmethod
    def _input_wrapper(x):
        """Helper method to admit PyTorch tensors."""
        if isinstance(x, torch.Tensor):
            return x.cpu().numpy(), x.device
        else:
            return x, None

    @staticmethod
    def _output_wrapper(x, torch_device=None):
        """Helper method to produce PyTorch tensors."""
        if torch_device is not None:
            return torch.from_numpy(x).to(torch_device)
        else:
            return x

    def __matmul__(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        Casts given PyTorch tensor into NumPy before applying matmul.
        Then casts produced result back into same PyTorch datatype and device.
        """
        x, device = self._input_wrapper(x)
        result = self._output_wrapper(super().__matmul__(x), device)
        return result

    def __rmatmul__(self, x):
        """Adjoint (left) matrix-vector multiplication ``x @ self``.

        Casts given PyTorch tensor into NumPy before applying matmul.
        Then casts produced result back into same PyTorch datatype and device.
        """
        x, device = self._input_wrapper(x)
        result = self._output_wrapper(super().__rmatmul__(x), device)
        return result

    def __repr__(self):
        """Returns a string in the form TorchLinOpWrapper<LinOp ...>."""
        wrapper = self.__class__.__name__
        result = f"{wrapper}<{super().__repr__()}>"
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for synthetic matrix utilities."""


import pytest
import torch

from skerch.synthmat import SynthMat
from skerch.utils import rademacher_flip

from . import exp_decay, poly_decay, rng_seeds, snr_lowrank_noise, torch_devices


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def f64_rtol():
    """Relative error tolerance for float64."""
    result = {torch.float64: 1e-10}
    return result


@pytest.fixture
def f32_rtol():
    """Relative error tolerance for float32."""
    result = {torch.float32: 1e-3}
    return result


@pytest.fixture
def dims_ranks_square(request):
    """Dimension and rank for square matrices."""
    result = [
        (1, 1),
        (10, 1),
        (100, 10),
        (1_000, 10),
        (1_000, 50),
    ]
    if request.config.getoption("--quick"):
        result = result[:3]
    return result


@pytest.fixture
def heights_widths_ranks_fat(request):
    """Shape and rank for non-square matrices."""
    result = [
        (1, 10, 1),
        (10, 100, 1),
        (100, 1_000, 10),
        (1_000, 10_000, 100),
    ]
    if request.config.getoption("--quick"):
        result = result[:2]
    return result


@pytest.fixture
def decay_ew_atol():
    """Absolute tolerance to determine that eigenvals are strictly decaying."""
    result = {torch.float64: 1e-7, torch.float32: 1e-4}
    return result


# ##############################################################################
# # SEED CONSISTENCY
# ##############################################################################
def test_seed_consistency(
    rng_seeds,
    torch_devices,
    f64_rtol,
    f32_rtol,
    heights_widths_ranks_fat,
    snr_lowrank_noise,
    exp_decay,
    poly_decay,
):
    """Seed consistency for synthetic random matrices.

    Tests that same seed and shape lead to same operator with same results,
    and different otherwise.
    """
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype, rtol in {**f64_rtol, **f32_rtol}.items():
                # symmetric tests
                for h, w, r in heights_widths_ranks_fat:
                    # lowrank+noise
                    for snr in snr_lowrank_noise:
                        mat1 = SynthMat.lowrank_noise(
                            shape=(h, h),  # L+N must be square
                            rank=r,
                            snr=snr,
                            seed=seed,
                            dtype=dtype,
                            device=device,
                        )
                        mat2 = SynthMat.lowrank_noise(
                            shape=(h, h),  # L+N must be square
                            rank=r,
                            snr=snr,
                            seed=seed + 1,
                            dtype=dtype,
                            device=device,
                        )
                        mat3 = SynthMat.lowrank_noise(
                            shape=(h, h),  # L+N must be square
                            rank=r,
                            snr=snr,
                            seed=seed,
                            dtype=dtype,
                            device=device,
                        )
                        assert torch.allclose(
                            mat1, mat3, rtol
                        ), "Same seed different matrix?"
                        assert (
                            mat1 != mat2
                        ).any(), f"Different seed same matrix?, {mat1, mat3}"
                    # exp decay
                    for dec in exp_decay:
                        for psd in (True, False):
                            mat1 = SynthMat.exp_decay(
                                shape=(h, w),
                                rank=r,
                                decay=dec,
                                symmetric=False,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            mat2 = SynthMat.exp_decay(
                                shape=(h, w),
                                rank=r,
                                decay=dec,
                                symmetric=False,
                                seed=seed + 1,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            mat3 = SynthMat.exp_decay(
                                shape=(h, w),
                                rank=r,
                                decay=dec,
                                symmetric=False,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            assert torch.allclose(
                                mat1, mat3, rtol
                            ), "Same seed different matrix?"
                            assert (
                                mat1 != mat2
                            ).any(), (
                                f"Different seed same matrix?, {mat1, mat3}"
                            )
                    # poly decay
                    for dec in poly_decay:
                        for psd in (True, False):
                            mat1 = SynthMat.poly_decay(
                                shape=(h, w),
                                rank=r,
                                decay=dec,
                                symmetric=False,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            mat2 = SynthMat.poly_decay(
                                shape=(h, w),
                                rank=r,
                                decay=dec,
                                symmetric=False,
                                seed=seed + 1,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            mat3 = SynthMat.poly_decay(
                                shape=(h, w),
                                rank=r,
                                decay=dec,
                                symmetric=False,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            assert torch.allclose(
                                mat1, mat3, rtol
                            ), "Same seed different matrix?"
                            assert (
                                mat1 != mat2
                            ).any(), (
                                f"Different seed same matrix?, {mat1, mat3}"
                            )


# ##############################################################################
# # SYMMETRIC
# ##############################################################################
def test_symmetric(  # noqa: C901  # ignore "is too complex"
    rng_seeds,
    torch_devices,
    f64_rtol,
    f32_rtol,
    dims_ranks_square,
    snr_lowrank_noise,
    exp_decay,
    poly_decay,
    decay_ew_atol,
):
    """Tests assumed properties for symmetric synthetic matrices.

    Creates square, symmetric synthetic matrices and tests that:

    * there are no NaNs
    * they are indeed symmetric
    * their diagonals/spectra are correct
    * if non-PSD, spectra have significantly negative entries and recovery is
      still correct
    """
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype, rtol in {**f64_rtol, **f32_rtol}.items():
                # symmetric tests
                for dim, r in dims_ranks_square:
                    # lowrank+noise
                    for snr in snr_lowrank_noise:
                        mat = SynthMat.lowrank_noise(
                            shape=(dim, dim),
                            rank=r,
                            snr=snr,
                            seed=seed,
                            dtype=dtype,
                            device=device,
                        )
                        assert not mat.isnan().any(), f"{mat, device, dtype}"
                        assert torch.allclose(
                            mat, mat.T, rtol=rtol
                        ), "Matrix not symmetric?"
                        assert all(
                            mat[range(r), range(r)] >= 1
                        ), "mat[:rank] is not >=1 for given rank?"
                    # exp decay
                    for dec in exp_decay:
                        for psd in (True, False):
                            mat = SynthMat.exp_decay(
                                shape=(dim, dim),
                                rank=r,
                                decay=dec,
                                symmetric=True,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            ew = torch.linalg.eigvalsh(mat).flip(0)  # desc ord
                            assert (
                                not mat.isnan().any()
                            ), f"{mat, device, dtype}"
                            assert torch.allclose(
                                mat, mat.T, rtol=rtol
                            ), "Matrix not symmetric?"
                            assert torch.allclose(
                                ew[:r], torch.ones_like(ew[:r]), rtol=rtol
                            ), "ew[:rank] should be == 1"
                            ew_dec = 10.0 ** -(
                                dec * torch.arange(1, dim - r + 1, dtype=dtype)
                            ).to(device)
                            if not psd:
                                # apply rademacher to the decay
                                rademacher_flip(
                                    ew_dec, seed=seed + 1, inplace=True
                                )
                                # sort recovered eigenvals by descending mag
                                _, perm = ew.abs().sort(descending=True)
                                ew = ew[perm]
                                # check that we actually have negatives
                                if ew_dec.numel() > 0:
                                    assert (
                                        ew_dec < -abs(decay_ew_atol[dtype])
                                    ).any(), (
                                        "non-PSD should have "
                                        + "negative eigvals!"
                                    )
                            assert torch.allclose(
                                ew[r:],
                                ew_dec,
                                rtol=rtol,
                                # added atol due to eigvalsh
                                atol=decay_ew_atol[dtype],
                            ), "Eigenval decay mismatch!"
                    # poly decay
                    for dec in poly_decay:
                        for psd in (True, False):
                            mat = SynthMat.poly_decay(
                                shape=(dim, dim),
                                rank=r,
                                decay=dec,
                                symmetric=True,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                psd=psd,
                            )
                            ew = torch.linalg.eigvalsh(mat).flip(0)  # desc ord
                            assert (
                                not mat.isnan().any()
                            ), f"{mat, device, dtype}"
                            assert torch.allclose(
                                mat, mat.T, rtol=rtol
                            ), "Matrix not symmetric?"
                            assert torch.allclose(
                                ew[:r], torch.ones_like(ew[:r]), rtol=rtol
                            ), "ew[:rank] should be == 1"
                            ew_dec = (
                                torch.arange(2, dim - r + 2, dtype=dtype)
                                ** (-float(dec))
                            ).to(device)
                            if not psd:
                                # apply rademacher to the decay
                                rademacher_flip(
                                    ew_dec, seed=seed + 1, inplace=True
                                )
                                # sort recovered eigenvals by descending mag
                                _, perm = ew.abs().sort(descending=True)
                                ew = ew[perm]
                                # check that we actually have negatives
                                if ew_dec.numel() > 0:
                                    assert (
                                        ew_dec < -abs(decay_ew_atol[dtype])
                                    ).any(), (
                                        "non-PSD should have "
                                        + "negative eigvals!"
                                    )
                            assert torch.allclose(
                                ew[r:],
                                ew_dec,
                                rtol=rtol,
                                # added atol due to eigvalsh
                                atol=decay_ew_atol[dtype],
                            ), "Eigenval decay mismatch!"


# ##############################################################################
# # ASYMMETRIC
# ##############################################################################
def test_asymmetric_nonsquare(
    rng_seeds,
    torch_devices,
    f64_rtol,
    f32_rtol,
    heights_widths_ranks_fat,
    snr_lowrank_noise,
    exp_decay,
    poly_decay,
    decay_ew_atol,
):
    """Tests assumed properties for asymmetric synthetic matrices.

    Create square, symmetric synthetic matrices and tests that:
    * there are no NaNs
    * their diagonals/spectra are correct

    .. note::

      Since lowrank+noise must be symmetric, it is omitted here.
    """
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype, rtol in {**f64_rtol, **f32_rtol}.items():
                # symmetric tests
                for h, w, r in heights_widths_ranks_fat:
                    min_dim = min(h, w)
                    # exp decay
                    for dec in exp_decay:
                        mat = SynthMat.exp_decay(
                            shape=(h, w),
                            rank=r,
                            decay=dec,
                            symmetric=False,
                            seed=seed,
                            dtype=dtype,
                            device=device,
                        )
                        ew = torch.linalg.svdvals(mat)  # desc order
                        assert not mat.isnan().any(), f"{mat, device, dtype}"
                        assert torch.allclose(
                            ew[:r], torch.ones_like(ew[:r]), rtol=rtol
                        ), f"ew[:rank] should be == 1, {ew}"
                        ew_dec = 10.0 ** -(
                            dec * torch.arange(1, min_dim - r + 1, dtype=dtype)
                        ).to(device)
                        assert torch.allclose(
                            ew[r:],
                            ew_dec,
                            rtol=rtol,
                            # added atol due to eigvalsh
                            atol=decay_ew_atol[dtype],
                        ), "Eigenval decay mismatch!"
                    # poly decay
                    for dec in poly_decay:
                        mat = SynthMat.poly_decay(
                            shape=(h, w),
                            rank=r,
                            decay=dec,
                            symmetric=False,
                            seed=seed,
                            dtype=dtype,
                            device=device,
                        )
                        ew = torch.linalg.svdvals(mat)  # desc order
                        assert not mat.isnan().any(), f"{mat, device, dtype}"
                        assert torch.allclose(
                            ew[:r], torch.ones_like(ew[:r]), rtol=rtol
                        ), "ew[:rank] should be == 1"
                        ew_dec = (
                            torch.arange(2, min_dim - r + 2, dtype=dtype)
                            ** (-float(dec))
                        ).to(device)
                        assert torch.allclose(
                            ew[r:],
                            ew_dec,
                            rtol=rtol,
                            # added atol due to eigvalsh
                            atol=decay_ew_atol[dtype],
                        ), "Eigenval decay mismatch!"



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for orthogonal projection linear operators."""


import pytest
import torch

from skerch.a_posteriori import a_posteriori_error
from skerch.linops import NegOrthProjLinOp, OrthProjLinOp
from skerch.utils import BadShapeError, gaussian_noise

from . import rng_seeds, torch_devices


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def heights_widths():
    """Shapes of matrices to QR decompose."""
    result = [
        (1, 1),
        (10, 10),
        (100, 100),
        (1_000, 100),
    ]
    return result


@pytest.fixture
def tall_heights_widths():
    """Shapes of matrices with ``h > w``."""
    result = [
        (10, 1),
        (100, 10),
        (1000, 100),
    ]
    return result


@pytest.fixture
def dtypes_atols():
    """Datatypes and their absolute error tolerances."""
    result = {torch.float64: 1e-10, torch.float32: 1e-5}
    return result


@pytest.fixture
def num_test_measurements():
    """Number of a-posteriori measurements."""
    return 50


# ##############################################################################
# #
# ##############################################################################
def test_proj_correctness(
    rng_seeds,
    torch_devices,
    heights_widths,
    dtypes_atols,
    num_test_measurements,
):
    """Test case for orthogonal projectors.

    Draws random, thin orthonormal matrices, constructs the corresponding
    orthogonal and neg-orthogonal projectors, and tests that:

    * Each projector is symmetric
    * Each projector is idempotent
    * Both projectors are orthogonal to each other
    * Adding both projectors yields identity
    """
    for h, w in heights_widths:
        assert h >= w, "This test doesn't need/admit fat matrices!"
    #
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype, atol in dtypes_atols.items():
                for h, w in heights_widths:
                    mat = gaussian_noise(
                        (h, w),
                        mean=0.0,
                        std=1.0,
                        seed=seed,
                        dtype=dtype,
                        device=device,
                    )
                    Q = torch.linalg.qr(mat)[0]
                    proj = OrthProjLinOp(Q)
                    negproj = NegOrthProjLinOp(Q)
                    #
                    meas_vectors = gaussian_noise(
                        (h, num_test_measurements),
                        mean=0.0,
                        std=1.0,
                        seed=seed + 123,
                        dtype=dtype,
                        device=device,
                    )
                    proj_meas = proj @ meas_vectors
                    negproj_meas = negproj @ meas_vectors
                    # test symmetry
                    assert torch.allclose(
                        proj_meas, (meas_vectors.T @ proj).T, atol=atol
                    ), "Projector is not symmetric!"
                    assert torch.allclose(
                        negproj_meas, (meas_vectors.T @ negproj).T, atol=atol
                    ), "Negative projector is not symmetric!"
                    # test idempotence
                    assert torch.allclose(
                        proj_meas, proj @ proj_meas, atol=atol
                    ), "Projector is not idempotent!"
                    assert torch.allclose(
                        negproj_meas, negproj @ negproj_meas, atol=atol
                    ), "Negative projector is not idempotent!"
                    # test orthogonality
                    assert (
                        abs(proj @ negproj_meas)
                    ).max() <= atol, (
                        "Projector and its negation are not orthogonal!"
                    )
                    # test identity
                    assert (
                        abs(proj_meas + negproj_meas - meas_vectors).max()
                        <= atol
                    ), "Projector and its negation are not identity!"


def test_proj_shapes(torch_devices, dtypes_atols, tall_heights_widths):
    """Test case for shape consistencies of projection linops.

    Check that
    * Bad ``Q`` shapes (fat) give bad shape error
    * Correct inputs can be both vectors and matrices
    """
    for device in torch_devices:
        for dtype in dtypes_atols.keys():
            for h, w in tall_heights_widths:
                """
                First try to create w,h and expect Badshapeerror

                then actually create, and feed a vec and a mat. should work.
                """
                # create mat. It doesn't need to be orthogonalized for this test
                mat = gaussian_noise(
                    (h, w),
                    mean=0.0,
                    std=1.0,
                    seed=12345,
                    dtype=dtype,
                    device=device,
                )
                # creating projectors with fat matrix should throw error
                with pytest.raises(BadShapeError):
                    OrthProjLinOp(mat.T)
                with pytest.raises(BadShapeError):
                    NegOrthProjLinOp(mat.T)
                # now create matrices and feed vectors and matrices on both
                # sides. No error should be thrown and shapes should match.
                test_vec = torch.ones(h, dtype=dtype, device=device)
                test_mat = torch.ones((h, 2), dtype=dtype, device=device)
                proj = OrthProjLinOp(mat)
                negproj = NegOrthProjLinOp(mat)
                #
                assert (proj @ test_vec).shape == (h,)
                assert (proj @ test_mat).shape == (h, 2)
                assert (test_vec.T @ proj).shape == (h,)
                assert (test_mat.T @ proj).shape == (2, h)
                #
                assert (negproj @ test_vec).shape == (h,)
                assert (negproj @ test_mat).shape == (h, 2)
                assert (test_vec.T @ negproj).shape == (h,)
                assert (test_mat.T @ negproj).shape == (2, h)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""General ``pytest`` utilities, like fixtures.

.. note::

  See `docs <https://docs.pytest.org/en/stable/explanation/fixtures.html>`_.
"""


import pytest
import torch


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def torch_devices():
    """Available PyTorch devices for the unit tests."""
    result = ["cpu"]
    if torch.cuda.is_available():
        result.append("cuda")
    return result


@pytest.fixture
def rng_seeds(request):
    """Random seeds for the unit tests.

    .. note::
     This implicitly uses code from :mod:`.conftest`.
    """
    result = request.config.getoption("--seeds")
    return result


@pytest.fixture
def snr_lowrank_noise():
    """SNR values for Lowrank+noise matrix. The larger, the more noise.

    .. note::
      See `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_
    """
    result = [1e-3, 1e-2, 1e-1, 1]
    return result


@pytest.fixture
def exp_decay():
    """Exponential decay values. The larger, the faster decay.

    .. note::
      See `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_
    """
    result = [0.5, 0.1, 0.01]
    return result


@pytest.fixture
def poly_decay():
    """Polynomial decay values. The larger, the faster decay.

    .. note::
      See `[TYUC2019, 7.3.1] <https://arxiv.org/abs/1902.08651>`_
    """
    result = [2, 1, 0.5]
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for Scrambled Subsampled Randomized Fourier Transform (SSRFT).

See `[TYUC2019, 3.2] <https://arxiv.org/abs/1902.08651>`_.
"""

import pytest
import torch

from skerch.ssrft import SSRFT
from skerch.utils import BadShapeError

from . import rng_seeds, torch_devices


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def f64_rtol():
    """Relative error tolerance for float64.

    .. note::

      Float32 tolerance goes to zero for larger shapes (which is where SSVD
      makes sense), so we only test f64. Use f32 at own risk!
    """
    result = {torch.float64: 1e-10}
    return result


@pytest.fixture
def square_shapes():
    """Test shapes for square matrices."""
    result = [
        (1, 1),
        (10, 10),
        (100, 100),
        (10_000, 10_000),
        (1_000_000, 1_000_000),
    ]
    return result


@pytest.fixture
def fat_shapes():
    """Test shapes for fat (wide) matrices."""
    result = [
        (1, 10),
        (10, 100),
        (100, 1_000),
        (10_000, 100_000),
        (1_000_000, 10_000_000),
    ]
    return result


# ##############################################################################
# # TESTS
# ##############################################################################
def test_no_nans(torch_devices, f64_rtol, rng_seeds, square_shapes):
    """Tests that SSRFT yields no NaNs."""
    for seed in rng_seeds:
        for h, w in square_shapes:
            ssrft = SSRFT((h, w), seed=seed)
            for device in torch_devices:
                for dtype, _rtol in f64_rtol.items():
                    x = torch.randn(w, dtype=dtype).to(device)
                    y = ssrft @ x
                    xx = y @ ssrft
                    #
                    assert not x.isnan().any(), f"{ssrft, device, dtype}"
                    assert not y.isnan().any(), f"{ssrft, device, dtype}"
                    assert not xx.isnan().any(), f"{ssrft, device, dtype}"


def test_invertible(torch_devices, f64_rtol, rng_seeds, square_shapes):
    """Invertibility/orthogonality of quare SSRFT.

    Tests that, when input and output dimensionality are the same, the SSRFT
    operator is orthogonal, i.e. we can recover the input exactly via an
    adjoint operation.

    Also tests that it works for mat-vec and mat-mat formats.
    """
    for seed in rng_seeds:
        for h, w in square_shapes:
            ssrft = SSRFT((h, w), seed=seed)
            for device in torch_devices:
                for dtype, rtol in f64_rtol.items():
                    # matvec
                    x = torch.randn(w, dtype=dtype).to(device)
                    y = ssrft @ x
                    xx = y @ ssrft
                    #
                    assert torch.allclose(
                        x, xx, rtol=rtol
                    ), f"MATVEC: {ssrft, device, dtype}"
                    # matmat
                    x = torch.randn((w, 2), dtype=dtype).to(device)
                    y = ssrft @ x
                    xx = (y.T @ ssrft).T
                    #
                    assert torch.allclose(
                        x, xx, rtol=rtol
                    ), f"MATMAT: {ssrft, device, dtype}"
                    # matmat-shape tests
                    assert len(y.shape) == 2
                    assert len(xx.shape) == 2
                    assert y.shape[-1] == 2
                    assert xx.shape[-1] == 2


def test_seed_consistency(torch_devices, f64_rtol, rng_seeds, square_shapes):
    """Seed consistency of SSRFT.

    Test that same seed and shape lead to same operator with same results,
    and different otherwise.
    """
    for seed in rng_seeds:
        for h, w in square_shapes:
            ssrft = SSRFT((h, w), seed=seed)
            ssrft_same = SSRFT((h, w), seed=seed)
            ssrft_diff = SSRFT((h, w), seed=seed + 1)
            for device in torch_devices:
                for dtype, _rtol in f64_rtol.items():
                    # matvec
                    x = torch.randn(w, dtype=dtype).to(device)
                    assert ((ssrft @ x) == (ssrft_same @ x)).all()
                    # here, dim=1 may indeed result in same output, since
                    # there are no permutations or index-pickings, so 50/50.
                    # therefore we ignore that case.
                    if x.numel() > 1:
                        assert ((ssrft @ x) != (ssrft_diff @ x)).any()


def test_device_consistency(torch_devices, f64_rtol, rng_seeds, square_shapes):
    """Seed consistency of SSRFT across different devices.

    Test that same seed and shape lead to same operator with same results,
    even when device is different.
    """
    for seed in rng_seeds:
        for h, w in square_shapes:
            ssrft = SSRFT((h, w), seed=seed)
            for dtype, rtol in f64_rtol.items():
                # apply SSRFT on given devices and check results are equal
                x = torch.randn(w, dtype=dtype)
                y = [(ssrft @ x.to(device)).cpu() for device in torch_devices]
                for yyy in y:
                    assert torch.allclose(
                        yyy, y[0], rtol=rtol
                    ), "SSRFT inconsistency among devices!"


def test_unsupported_tall_ssrft(rng_seeds, fat_shapes):
    """Tail SSRFT linops are not supported."""
    for seed in rng_seeds:
        for h, w in fat_shapes:
            with pytest.raises(BadShapeError):
                # If this line throws a BadShapeError, the test passes
                SSRFT((w, h), seed=seed)


def test_input_shape_mismatch(rng_seeds, fat_shapes, torch_devices, f64_rtol):
    """Test case for SSRFT shape consistency."""
    for seed in rng_seeds:
        for h, w in fat_shapes:
            ssrft = SSRFT((h, w), seed=seed)
            for device in torch_devices:
                for dtype, _rtol in f64_rtol.items():
                    # forward matmul
                    x = torch.empty(w + 1, dtype=dtype).to(device)
                    with pytest.raises(BadShapeError):
                        ssrft @ x
                    # adjoint matmul
                    x = torch.empty(h + 1, dtype=dtype).to(device)
                    with pytest.raises(BadShapeError):
                        x @ ssrft



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Functionality to support sketched estimation of upper/lower triangulars.

Given an arbitrary square ``linop``, this module implements matrix-free
functionality to compute left- and right matrix-multiplications of both
``lower_triang(linop)`` and ``upper_triang(linop)``.

This is done by first exactly computing a limited number of "main rectangles",
a.k.a. "steps", that fully belong to the triangular submatrix, and then
estimating the "serrated pattern" that they leave as residual via modified
Hutchinson. The addition of the steps plus the serrated pattern forms the full
triangle. See :class:`TriangularLinOp` and the
:ref:`Examples section<Examples>` for more details.

Contents:

* Hadamard pattern to estimate block-triangular matrix-vector products using
  Hutchinson's method (used by :class:`TriangularLinOp`).
* The :class:`TriangularLinOp` linear operator implementing triangular
  matrix-vector products for arbitrary square linops.
"""


import torch

from .linops import BaseLinOp
from .ssrft import SSRFT
from .subdiagonals import subdiag_hadamard_pattern
from .utils import BadShapeError


# ##############################################################################
# # HADAMARD
# ##############################################################################
def serrated_hadamard_pattern(
    v, blocksize, with_main_diagonal=True, lower=True, use_fft=False
):
    """Shifted copies of vectors for block-triangular Hutchinson estimation.

    :param v: Torch vector expected to contain zero-mean, uncorrelated entries.
    :param with_main_diagonal: If true, the main diagonal will be included
          in the patterns, otherwise excluded.
    :param lower: If true, the block-triangles will be below the diagonal,
      otherwise above.
    :param use_fft: See :func:`subdiag_hadamard_pattern`.
    :returns: A vector of same shape, type and device as ``v``, composed of
      shifted copies of ``v`` following a block-triangular pattern.

    For example, given a 10-dimensional vector, the corresponding serrated
    pattern with ``blocksize=3, with_main_diagonal=True, lower=True`` yields
    the following entries:

    * ``v1``
    * ``v1 + v2``
    * ``v1 + v2 + v3``
    * ``v4``
    * ``v4 + v5``
    * ``v4 + v5 + v6``
    * ``v7``
    * ``v7 + v8``
    * ``v7 + v8 + v9``
    * ``v10``

    If main diagonal is excluded, it will look like this instead:

    * ``0``
    * ``v1``
    * ``v1 + v2``
    * ``0``
    * ``v4``
    * ``v4 + v5``
    * ``0``
    * ``v7``
    * ``v7 + v8``
    * ``0``

    And if ``lower=False``, it will look like this instead:

    * ``v1``
    * ``v4 + v3 + v2``
    * ``v4 + v3``
    * ``v4``
    * ``v7 + v6 + v5``
    * ``v7 + v6``
    * ``v7``
    * ``v10 + v9 + v8``
    * ``v10 + v9``
    * ``v10``
    """
    if blocksize < 1:
        raise ValueError("Block size must be a positive scalar!")
    #
    len_v = len(v)
    if use_fft:
        if lower:
            idxs = range(len_v) if with_main_diagonal else range(1, len_v)
            result = subdiag_hadamard_pattern(v, idxs, use_fft=True)
            for i in range(0, len_v, blocksize):
                mark = i + blocksize
                offset = sum(v[i:mark])
                result[mark:] -= offset
        else:
            idxs = (
                range(0, -len_v, -1)
                if with_main_diagonal
                else range(-1, -len_v, -1)
            )
            result = subdiag_hadamard_pattern(v, idxs, use_fft=True)
            for i in range(0, len_v, blocksize):
                mark = len_v - (i + blocksize)
                offset = sum(v[mark : (mark + blocksize)])
                result[:mark] -= offset
    else:
        if with_main_diagonal:
            result = v.clone()
        else:
            result = torch.zeros_like(v)
        #
        for i in range(len_v - 1):
            block_n, block_i = divmod(i + 1, blocksize)
            if block_i == 0:
                continue
            # get indices for result[out_beg:out_end] = v[beg:end]
            if lower:
                beg = block_n * blocksize
                end = beg + block_i
                out_end = min(beg + blocksize, len_v)
                out_beg = out_end - block_i
            else:
                end = len_v - (block_n * blocksize)
                beg = end - block_i
                out_beg = max(0, end - blocksize)
                out_end = out_beg + block_i
            # add to serrated pattern
            result[out_beg:out_end] += v[beg:end]
    #
    return result


# ##############################################################################
# # TRIANGULAR LINOP
# ##############################################################################
class TriangularLinOp(BaseLinOp):
    r"""Given a square linop, compute products with one of its triangles.

    The triangle of a linear operator can be approximated from the full operator
    via a "staircase pattern" of exact measurements, whose computation is exact
    and fast. For example, given an operator of shape ``(1000, 1000)``, and
    stairs of size 100, yields 9 exact measurements strictly under the diagonal,
    the first one covering ``lop[100:, :100]``, the next one
    ``lop[200, 100:200]``, and so on. The more measurements, the more closely
    the full triangle is approximated.

    Note that this staircase pattern leaves a block-triangular section of the
    linop untouched (near the main diagonal). This part can be then estimated
    with the help of  :func:`serrated_hadamard_pattern`, completing the
    triangular approximation, as follows:


    Given a square linear operator :math:`A`, and random vectors
    :math:`v \sim \mathcal{R}` with :math:`\mathbb{E}[v v^T] = I`, consider
    the generalized Hutchinson diagonal estimator:

    .. math::

      f(A) =
      \mathbb{E}_{v \sim \mathcal{R}} \big[ \varphi(v) \odot Av \big]

    In this case, if the :math:`\varphi` function follows a "serrated
    Hadamard pattern", :math:`f(A)` will equal a block-triangular subset of
    :math:`A`.
    """

    def __init__(
        self,
        lop,
        stair_width=None,
        num_hutch_measurements=0,
        lower=True,
        with_main_diagonal=True,
        seed=0b1110101001010101011,
        use_fft=True,
    ):
        """Init method.

        :param lop: A square linear operator of order ``dims``, such that
          ``self @ v`` will equal ``triangle(lop) @ v``. It must implement a
          ``lop.shape = (dims, dims)`` attribute as well as the left- and right-
          matmul operator ``@``, interacting with torch tensors.
        :param stair_width: Width of each step in the staircase pattern. If
          it is 1, a total of ``dims`` exact measurements will be performed.
          If it equals ``dims``, no exact measurements will be performed (since
          the staircase pattern would cover the full triangle). The step size
          regulates this trade-off: Ideally, we want as many exact measurements
          as possible, but not too many. If no value is provided, ``dims // 2``
          is chosen by default, such that only 1 exact measurement is performed.
        :param num_hutch_measurements: The leftover entries from the
          staircase measurements are approximated here using an extension of
          the Hutchinson diagonal estimator. This estimator generally requires
          many measurements to be informative, and it can even be counter-
          productive if not enough measurements are given. If ``lop``is not
          diagonally dominant, consider setting this to 0 for a sufficiently
          good approximation via ``staircase_measurements``. Otherwise,
          make sure to provide a sufficiently high number of measurements.
        :param lower: If true, lower triangular matmuls will be computed.
          Otherwise, upper triangular.
        :param with_main_diagonal: If true, the main diagonal will be included
          in the triangle, otherwise excluded. If you already have precomuted
          the diagonal elsewhere, consider excluding it from this approximation,
          and adding it separately.
        :param seed: Seed for the random SSRFT measurements used in the
          Hutchinson estimator.
        :param use_fft: Whether to use FFT for the Hutchinson estimation. See
          :func:`subdiag_hadamard_pattern` for more details.
        """
        h, w = lop.shape
        if h != w:
            raise BadShapeError("Only square linear operators supported!")
        self.dims = h
        if self.dims < 1:
            raise BadShapeError("Empty linear operators not supported!")
        if stair_width is None:
            stair_width = max(1, self.dims // 2)
        assert stair_width > 0, "Stair width must be a positive int!"
        self.lop = lop
        self.use_fft = use_fft
        self.lower = lower
        self.with_main = with_main_diagonal
        self.n_hutch = num_hutch_measurements
        #
        self.stair_width = stair_width
        num_stair_meas = sum(
            1 for _ in self._iter_stairs(self.dims, stair_width)
        )
        self.quack = num_stair_meas
        assert (
            num_stair_meas <= self.dims
        ), "More staircase measurements than dimensions??"
        assert (
            self.n_hutch <= self.dims
        ), "More Hutchinson measurements than dimensions??"
        assert (
            num_stair_meas + self.n_hutch
        ) <= self.dims, "More total measurements than dimensions??"
        #
        self.ssrft = SSRFT((self.n_hutch, self.dims), seed=seed)
        super().__init__(lop.shape)  # this sets self.shape also

    @staticmethod
    def _iter_stairs(dims, stair_width):
        """Helper method to iterate over staircase indices.

        This method implements an iterator that yields ``(begin, end)`` index
        pairs for each staircase-pattern step. It terminates before ``end``
        is equal or greater than ``self.dims``, since only full steps are
        considered.
        """
        beg, end = 0, stair_width
        while end < dims:
            yield (beg, end)
            beg = end
            end = beg + stair_width

    def _matmul_helper(self, x, adjoint=False):
        """Forward and adjoint triangular matrix multiplications.

        Since forward and adjoint matmul share many common computations, this
        method implements both at the same time. The specific mode can be
        dispatched using the ``adjoint`` parameter.
        """
        self.check_input(x, adjoint=adjoint)
        # we don't factorize this method because we want to share buff
        # across both loops to hopefully save memory
        buff = torch.zeros_like(x)
        result = torch.zeros_like(x)
        # add step computations to result
        for beg, end in self._iter_stairs(self.dims, self.stair_width):
            if (not adjoint) and self.lower:
                buff[beg:end] = x[beg:end]
                result[end:] += (self.lop @ buff)[end:]
                buff[beg:end] = 0
            elif (not adjoint) and (not self.lower):
                buff[end:] = x[end:]
                result[beg:end] += (self.lop @ buff)[beg:end]
                buff[end:] = 0
            elif adjoint and self.lower:
                buff[end:] = x[end:]
                result[beg:end] += (buff @ self.lop)[beg:end]
                buff[end:] = 0
            elif adjoint and (not self.lower):
                buff[beg:end] = x[beg:end]
                result[end:] += (buff @ self.lop)[end:]
                buff[beg:end] = 0
            else:
                raise RuntimeError("This should never happen")
        # add Hutchinson estimator to result
        for i in range(self.n_hutch):
            buff[:] = self.ssrft.get_row(i, x.dtype, x.device)
            result[:] += serrated_hadamard_pattern(
                buff,
                self.stair_width,
                self.with_main,
                lower=(self.lower ^ adjoint),  # pattern also depends on adj
                use_fft=self.use_fft,
            ) * (
                ((buff * x) @ self.lop) if adjoint else (self.lop @ (buff * x))
            )
        return result

    def __matmul__(self, x):
        """Forward (right) matrix-vector multiplication ``self @ x``.

        See parent class for more details.
        """
        return self._matmul_helper(x, adjoint=False)

    def __rmatmul__(self, x):
        """Adjoint (left) matrix-vector multiplication ``x @ self``.

        See parent class for more details.
        """
        return self._matmul_helper(x, adjoint=True)

    def __repr__(self):
        """Readable string version of this object.

        Returns a string in the form
        <TriangularlLinOp[lop](lower/upper)>.
        """
        clsname = self.__class__.__name__
        lopstr = self.lop.__repr__()
        lower_str = "lower" if self.lower else "upper"
        diag_str = "with" if self.with_main else "no"
        result = f"<{clsname}[{lopstr}]({lower_str}, {diag_str} main diag)>"
        return result



# -*- coding: utf-8 -*-

r"""Command Line Interface
==========================

This example demonstrates the functionality in ``skerch`` that can be directly
accessed from the command line interface (CLI).

We simply assume that ``skerch`` is accessible to the available ``python``
interpreter.

The following imports are only needed to exemplify the Python code:
"""


import os
from tempfile import TemporaryDirectory

from skerch.__main__ import main_wrapper as skerch_main

# %%
#
# ##############################################################################
#
# CLI help documentation
# ----------------------
#
# The following command provides information about positional and optional
# CLI arguments::
#
#   python -m skerch -h
#
# The following Python code is equivalent:


skerch_main(["-h"])


# %%
#
# ##############################################################################
#
# A-priori hyperparameter estimation
# ----------------------------------
#
# Sketched decompositions have hyperparameters for the number of inner and outer
# measurements to be performed. A natural question is, given a limited amount
# of memory (provided in number of matrix entries), how many inner and outer
# measurements should we perform to get the best results?
#
# For asymmetric matrices, this can be quickly checked as follows::
#
#   python -m skerch prio_hpars --shape=100,200 --budget=12345
#
# The following Python code is equivalent:

skerch_main(["prio_hpars", "--shape=100,200", "--budget=1234"])

# %%
# The ``[CHECK]`` field shows that the memory requirements are optimally
# satisfied for the given solution and constraints (see respective
# documentation for more details)

# %%
#
# ##############################################################################
#
# A-posteriori error bounds
# -------------------------
#
# Once a linear operator is approximated via sketched methods, it is possible
# to efficiently estimate how good of an approximation it is (see respective
# documentation for more details).
#
# The quality of this estimation depends only on the number of *a-posteriori*
# measurements performed and it can be quickly checked as follows::
#
#   python -m skerch post_bounds --apost_n=30 --apost_err=0.5
#
# The following Python code is equivalent:

skerch_main(["post_bounds", "--apost_n=30", "--apost_err=0.75"])

# %%
# We see in this example that, if we performed 30 measurements and we got an
# error estimate of :math:`\hat{\varepsilon}`, the probability of the *actual*
# error :math:`\varepsilon` being outside of the
# :math:`(0.25\hat{\varepsilon}, 1.75\hat{\varepsilon})` range is provided.


# %%
#
# ##############################################################################
#
# Creating HDF5 layout for distributed sketches
# ---------------------------------------------
#
# HDF5 files allow to efficiently read and write large numerical arrays in
# an out-of-core, distributed fashion. This is useful to perform sketched
# decompositions of (very) large linear operators, since both storage
# and measurements can be distributed (see respective documentation for details
# on how to work with these files using ``skerch``.
#
# A HDF5 layout can be conveniently created as follows::
#
#   python -m skerch create_hdf5_layout --hdf5dir=/tmp/test_skerch \
#          --dtype=float64 --shape=123,234 --outer=30 --inner=60
#
# The following Python code is equivalent (up to the use of ``tmpdir``):

tmpdir = TemporaryDirectory()
skerch_main(
    [
        "create_hdf5_layout",
        f"--hdf5dir={tmpdir.name}",
        "--dtype=float64",
        "--shape=123,234",
        "--outer=30",
        "--inner=60",
    ]
)

# %%
# The script prints the paths and main properties of the generated HDF5 files,
# which comprise several
# `HDF5 virtual datasets <https://docs.h5py.org/en/stable/vds.html>`_ which act
# as layouts to hold the left-outer, right-outer and inner measurements.
# In this example, a sketched decomposition for a ``torch.float64`` operator
# of height 123 and width 234 is being prepared, comprising 30 outer and 60
# inner measurements (which can be good if e.g. the effective rank of the
# operator is around 10).
#
# If the ``--sym`` flag is provided, ``rightouter`` files won't be created.


# %%
#
# ##############################################################################
#
# Merging distributed HDF5 sketches
# ---------------------------------
#
# Although decentralized measurement and storage via HDF5 virtual datasets has
# many advantages, some operations may require to process the measurements in
# a centralized fashion. For instance, many operative systems do not allow a
# single process to keep thousands of files open at the same time.
#
# A solution is to merge all individual HDF5 files from the virtual dataset
# into a single, centralized HDF5 file of the same size. The following command
# merges the previously created left-outer measurement layout::
#
#   python -m skerch merge_hdf5 --delete_subfiles --ok_flag=initialized \
#          --in_path /tmp/tmp4fswvvk2/leftouter_ALL.h5
#
# The following Python code is equivalent (up to the use of ``tmpdir``):

skerch_main(
    [
        "merge_hdf5",
        "--delete_subfiles",
        "--ok_flag=initialized",
        "--in_path",
        os.path.join(tmpdir.name, "leftouter_ALL.h5"),
    ]
)
tmpdir.cleanup()

# %%
# Note the following:
#
# * If the ``--delete_subfiles`` flag is provided, each individual distributed
#   HDF5 file will be deleted after being merged.
# * If ``--ok_flag`` is provided, the script will check that all HDF5 flags
#   match this value before proceeding. This can be used to ensure that all
#   distributed measurements have been performed before merging/deleting.
# * The ``--out_path`` flag can also be provided to set the location of the
#   merged HDF5 file. If none provided, the path of the former virtual dataset
#   is used. In any case the path of the merged dataset is returned.



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""In-core, single-process sketched decompositions.

.. note::
  Often, when the decomposition is small enough to fit into a single process
  and core, the traditional and exact algorithms may be preferred.
  For the cases that do not fit, see :mod:`.distributed_decompositions`.
"""


import torch

from .distributed_decompositions import (
    orthogonalize,
    solve_core_seigh,
    solve_core_ssvd,
)
from .distributed_measurements import innermeas_idx_torch, ssrft_idx_torch

#
from .ssrft import SSRFT


# ##############################################################################
# # TRUNCATE CORE MATRIX
# ##############################################################################
def truncate_core(k, core_U, core_S, core_Vt=None):
    """Truncation of core sketches.

    Truncates core matrices as returned by :func:`ssvd` or :func:`seigh` down to
    rank ``k``. In SEIGH, ``core_Vt`` is the same as ``core_U.T``, so it can be
    omitted.

    :param int k: Truncation rank of the output core elements.
    :param core_U: Left core basis matrix of shape ``(outer_dim, outer_dim)``.
    :param core_S: Core singular values of shape ``(outer_dim,)``.
    :param core_Vt: If given, right core matrix analogous to ``core_U``.
    :returns: Truncated ``(core_U, core_S, core_Vt)`` such that ``core_S``
      retains the first ``k`` entries, and the core matrices are truncated
      accordingly. If ``core_Vt`` is none, only the first two elements are
      returned.
    """
    core_U = core_U[:, :k]
    core_S = core_S[:k]
    if core_Vt is not None:
        core_Vt = core_Vt[:k, :]
        return core_U, core_S, core_Vt
    else:
        return core_U, core_S


# ##############################################################################
# # SKETCHED SVD
# ##############################################################################
def ssvd(
    op,
    op_device,
    op_dtype,
    outer_dim,
    inner_dim,
    seed=0b1110101001010101011,
):
    """Sketched Singular Value Decomposition (SVD).

    Sketched SVD of any given linear operator, as introduced in
    `[TYUC2019, 2] <https://arxiv.org/abs/1902.08651>`_. As with any SVD, the
    operator does not have to be square, symmetric, or PSD.

    :param op: Linear operator that implements the left- and right matmul
      operators via ``@``, as well as the ``shape=(h, w)`` attribute, but it
      doesn't need to be in explicit matrix form.
    :param int outer_dim: Number of outer measurements to be performed.
    :param int inner_dim: Number of inner measurements to be performed. It
      should be larger than ``outer_dim``, typically twice is reccommended.
    :returns: ``(Q, U, S, V.T, P.T)``, so ``Q @ U @ diag(S) @ V.T @ P.T`` is a
      low-rank approximation of ``op``. This is an SVD since ``S`` contains the
      non-negative singular values (in non-ascending order), and ``Q, U``
      contain orthogonal columns with ``Q.shape = (h, outer_dim)``, and
      ``U.shape = (outer_dim, outer_dim)``.
    """
    # convenience params
    h, w = op.shape
    lo_seed, ro_seed, li_seed, ri_seed = seed, seed + 1, seed + 2, seed + 3
    # perform random left outer measurements and their QR
    lo_t = torch.empty((outer_dim, w), dtype=op_dtype).to(op_device)
    for i in range(outer_dim):
        ssrft_idx_torch(
            i,
            outer_dim,
            op,
            op_device,
            op_dtype,
            lo_t[i],
            seed=lo_seed,
            adjoint=True,
        )
    orthogonalize(lo_t.T, overwrite=True)
    # perform random right outer measurements and their QR
    ro = torch.empty((h, outer_dim), dtype=op_dtype).to(op_device)
    for i in range(outer_dim):
        ssrft_idx_torch(
            i,
            outer_dim,
            op,
            op_device,
            op_dtype,
            ro[:, i],
            seed=ro_seed,
            adjoint=False,
        )
    orthogonalize(ro, overwrite=True)
    # perform random inner measurements
    inner_meas = torch.empty((inner_dim, inner_dim), dtype=op_dtype).to(
        op_device
    )
    for i in range(inner_dim):
        innermeas_idx_torch(
            i,
            inner_dim,
            op,
            op_device,
            op_dtype,
            inner_meas[:, i],
            li_seed,
            ri_seed,
        )
    # Solve core op to yield initial approximation
    left_inner_ssrft = SSRFT((inner_dim, h), seed=li_seed)
    right_inner_ssrft = SSRFT((inner_dim, w), seed=ri_seed)
    core_U, core_S, core_Vt = solve_core_ssvd(
        ro, lo_t.T, inner_meas, left_inner_ssrft, right_inner_ssrft
    )
    #
    return ro, core_U, core_S, core_Vt, lo_t


# ##############################################################################
# # SKETCHED EIGH
# ##############################################################################
def seigh(
    op,
    op_device,
    op_dtype,
    outer_dim,
    inner_dim,
    seed=0b1110101001010101011,
):
    """Sketched Hermitian Eigendecomposition (EIGH).

    Sketched EIGH of any given linear operator, modified from the SSVD from
    `[TYUC2019, 2] <https://arxiv.org/abs/1902.08651>`_. It leverages Hermitian
    symmetry of the operator to require half the measurements, memory and
    arithmetic. The operator must be square and Hermitian, but doesn't have to
    be PSD.

    :param op: Hermitian linear operator that implements the left- and right
      matmul operators via ``@``, as well as the ``shape=(d, d)`` attribute,
      but it doesn't need to be in explicit matrix form.
    :param int outer_dim: Number of outer measurements to be performed.
    :param int inner_dim: Number of inner measurements to be performed. It
      should be larger than ``outer_dim``, typically twice is reccommended.
    :returns: ``(Q, U, S)``, so ``Q @ U @ diag(S) @ U.T @ Q.T`` is a
      low-rank approximation of ``op``. This is an EIGH since ``S`` contains the
      eigenvalues (in descending magnitude), and ``Q, U``
      contain orthogonal columns with ``Q.shape = (h, outer_dim)``, and
      ``U.shape = (outer_dim, outer_dim)``.
    """
    assert (
        inner_dim >= outer_dim
    ), "Can't have more outer than inner measurements!"
    # convenience params and square-matrix check
    li_seed, ri_seed = seed, seed + 1
    h, w = op.shape
    assert h == w, f"Seigh expects square matrices! {(h, w)}"
    # perform inner (and recycled outer) random measurements
    outer = torch.empty((h, outer_dim), dtype=op_dtype).to(op_device)
    inner = torch.empty((inner_dim, inner_dim), dtype=op_dtype).to(op_device)
    for i in range(inner_dim):
        out_buff = innermeas_idx_torch(
            i,
            inner_dim,
            op,
            op_device,
            op_dtype,
            inner[:, i],
            li_seed,
            ri_seed,
        )
        if i < outer_dim:
            outer[:, i] = out_buff
    # orthogonalize outer measurements
    orthogonalize(outer, overwrite=True)
    # Solve core op to yield initial approximation
    li_ssrft = SSRFT((inner_dim, h), seed=li_seed)
    ri_ssrft = SSRFT((inner_dim, w), seed=ri_seed)
    core_U, core_S = solve_core_seigh(outer, inner, li_ssrft, ri_ssrft)
    #
    return outer, core_U, core_S



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytests distributed sketched EIGH utils."""


import torch

from skerch.distributed_decompositions import (
    DistributedHDF5,
    create_hdf5_layout,
    orthogonalize,
    solve_core_seigh,
)
from skerch.distributed_measurements import innermeas_idx_torch
from skerch.ssrft import SSRFT


# ##############################################################################
# # MOCK DISTRIBUTED DECOMPOSITIONS
# ##############################################################################
def mock_dseigh(dirname, lop, lop_device, lop_dtype, outer, inner, seed):
    """Mock implementation of a distributed sketched EIGH.

    This function emulates, in a single process, the steps that would be done
    in a distributed system to compute a sketched EIGH.
    """
    assert inner >= outer, "Can't have more outer than inner measurements!"
    processing, success = "0", "1"  # arbitrary flags
    # convenience params
    h, w = lop.shape
    assert h == w, "Square matrix expected!"
    # outer_seed, inner_seed = seed, seed + 1
    li_seed, ri_seed = seed, seed + 1
    # create HDF5 layout (centralized)
    (
        (outer_path, outer_subpaths),
        (_, _),
        (inner_path, inner_subpaths),
    ) = create_hdf5_layout(
        dirname,
        (h, w),
        lop_dtype,
        outer,
        inner,
        lo_fmt="leftouter_{}.h5",
        ro_fmt="rightouter_{}.h5",
        inner_fmt="core_{}.h5",
        with_ro=False,
    )
    # perform inner (and recycled outer) random measurements (distributed)
    for i in range(inner):
        inn_vals, inn_flag, inn_h5 = DistributedHDF5.load(inner_subpaths[i])
        out_buff = innermeas_idx_torch(
            i,
            inner,
            lop,
            lop_device,
            lop_dtype,
            inn_vals,
            li_seed,
            ri_seed,
            flag=inn_flag,
            processing_flag=processing,
            success_flag=success,
        )
        # check successful flags
        assert inn_flag[0].decode() == success, f"Bad inner measurement? {i}"
        inn_h5.close()
        # recycle outer measurements
        if i < outer:
            out_vals, out_flag, out_h5 = DistributedHDF5.load(outer_subpaths[i])
            out_vals[:] = out_buff
            out_flag[0] = success
            assert (
                out_flag[0].decode() == success
            ), f"Bad outer measurement? {i}"
            out_h5.close()
    del out_buff
    # merge virtual HDF5 datasets into monolithic without memory overhead
    DistributedHDF5.merge_all(
        outer_path,
        out_path=None,
        delete_subfiles_while_merging=True,
        check_success_flag=success,
    )
    DistributedHDF5.merge_all(
        inner_path,
        out_path=None,
        delete_subfiles_while_merging=True,
        check_success_flag=success,
    )
    # load all components and check that all flags are OK (centralized)
    outr, outr_flags, o_h5 = DistributedHDF5.load(outer_path, filemode="r+")
    innr, innr_flags, i_h5 = DistributedHDF5.load(inner_path, filemode="r+")
    assert all(f.decode() == "1" for f in outr_flags), "Bad outer flags!"
    assert all(f.decode() == "1" for f in innr_flags), "Bad inner flags!"
    # orthogonalize outer measurements (in-place)
    orthogonalize(outr, overwrite=True)
    # Solve core op and decompose via eigh
    li_ssrft = SSRFT((inner, h), seed=li_seed)
    ri_ssrft = SSRFT((inner, w), seed=ri_seed)
    core_U, core_S = solve_core_seigh(outr, innr, li_ssrft, ri_ssrft)
    # for this utest, convert everything to explicit torch tensors
    outr = torch.from_numpy(outr[:]).to(lop_device)
    core_S = torch.from_numpy(core_S).to(lop_device)
    core_U = torch.from_numpy(core_U).to(lop_device)
    # close HDF5 files and return
    o_h5.close()
    i_h5.close()
    return outr, core_U, core_S
